pro prepare_blocks_artemis, file_param_blocks, ind_arrays_eff, orient, nt, nblocks, xdeg_blocks, ydeg_blocks, $
                    narr=narr, nscans=nscans, nbolo=nbolo, boloref=boloref, arrays_default=arrays_default, $
                    max_n_samples=max_n_samples, ind_arrays=ind_arrays, debug=debug, $
                    dir_in=dir_in, files_scans=files_scans, hdr_ref=hdr_ref, cutout=cutout, $
                    file_XYscan=file_XYscan, scanrot_optim=scanrot_optim, vscan_status, dt_status, $
                    flag_exit=flag_exit, list_scanrot=list_scanrot, $
                    offset_ra_as=offset_ra_as, offset_dec_as=offset_dec_as, spiral=spiral



;;; prepares coordinates frame

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



crs = string(13b) + string(10b)

cut_margin = 0.05
ind_order_scans = indgen(nscans)

nf = n_elements(files_scans)
restore, dir_in + files_scans(0)

nt = n_elements(scan.time_s)
if !version.memory_bits ge 64 then nt = long64(nt)

tags = tag_names(scan)
pos = stregex(tags, 'vscan_status', /fold_case)
w = where(pos ge 0, cw)
if cw ge 1 then vscan_status = scan.(w(0)) else warn_vscan = 1
pos = stregex(tags, 'dt_status', /fold_case)
w = where(pos ge 0, cw)
if cw ge 1 then dt_status = scan.(w(0)) else warn_dt = 1


;;; check whether data have to be sliced into blocks:
;;; (in which case only one detector is processed, with a scan-parallel orientation)

ind_arrays_eff = ind_arrays
w = where(ind_arrays eq 1, cw)
ntb = nt * nbolo(w(0)) * nscans
print, crs + 'nt * nb ~ ' + chain(ntb)

if ((ntb gt max_n_samples or nblocks gt 1) and (not keyword_set(cutout))) then begin
   print, crs + '*******************************************************************************'
   print, crs + 'The ' + arrays_default(w(0)) + ' map is to be sliced into overlapping sub-maps.'
   if nblocks le 1 then print, ' (to keep the processed data volume within reasonable bounds)'
   if orient eq 'astro' then begin
      print, crs + 'Changing the map orientation to process the blocks.'
      orient = 'scan'
   endif
   print, crs + '*******************************************************************************'
   print, crs
endif



;;; coordinate vectors of all bolometers in degrees (Xra = -RA, Ydec = DEC):

Xra_bolos = ptrarr(narr, /allocate_heap)
Ydec_bolos = ptrarr(narr, /allocate_heap)
mid_Xra = dblarr(narr)
mid_Ydec = dblarr(narr)
off_cen_Xra = dblarr(narr, nf)
off_cen_Ydec = dblarr(narr, nf)
drdeg_max = dblarr(narr)
max_xdeg = dblarr(narr, nf)
max_ydeg = dblarr(narr, nf)
scanrot_deg = 0
tab_mid_x = dblarr(nf)
tab_mid_y = dblarr(nf)

if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then begin
   p_Xra_bolos = ptrarr(narr, /allocate_heap)
   p_Ydec_bolos = ptrarr(narr, /allocate_heap)
   p_mid_Xra = dblarr(narr)
   p_mid_Ydec = dblarr(narr)
   p_off_cen_Xra = dblarr(narr, nf)
   p_off_cen_Ydec = dblarr(narr, nf)
   p_max_xdeg = dblarr(narr, nf)
   p_max_ydeg = dblarr(narr, nf)
endif


list_scanrot = dblarr(nf)
if (not keyword_set(scanrot_optim)) then scanrot_optim = !values.f_nan

re_read_scans:


for detarr = 0, narr - 1 do if ind_arrays(detarr) eq 1 then begin
   pos = stregex(tags, 'Xra_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Xra_bolos_scan = scan.(w(0))
   pos = stregex(tags, 'Ydec_bolos_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0)
   Ydec_bolos_scan = scan.(w(0))

   if keyword_set(cutout) then begin
      mid_Xra(detarr) = -double(cutout(0))
      mid_Ydec(detarr) = double(cutout(1))
   endif else begin
      mid_Xra(detarr) = median(Xra_bolos_scan)
      mid_Ydec(detarr) = median(Ydec_bolos_scan)
      tab_mid_x(0) = mid_Xra(detarr)
      tab_mid_y(0) = mid_Ydec(detarr)
   endelse
   refx = mid_Xra(detarr)
   refy = mid_Ydec(detarr)

   gnomonic, Xra_bolos_scan, Ydec_bolos_scan, refx=refx, refy=refy

   if keyword_set(offset_ra_as) then begin
      Xra_bolos_scan -= (offset_ra_as(0) / 3600.D0)
      Ydec_bolos_scan += (offset_dec_as(0) / 3600.D0)
   endif

;;;
   wb = where(total(finite(Xra_bolos_scan), 1) ge 0.9 * max(total(finite(Xra_bolos_scan), 1)) $
      and total(finite(Ydec_bolos_scan), 1) ge 0.9 * max(total(finite(Ydec_bolos_scan), 1)))
   m = min((Xra_bolos_scan(nt/2L, wb) - median(Xra_bolos_scan(nt/2L, wb), /even))^2. $
      + (Ydec_bolos_scan(nt/2L, wb) - median(Ydec_bolos_scan(nt/2L, wb), /even))^2., bref)
   boloref(detarr) = wb(bref)
;;;
;;;

   if keyword_set(cutout) then begin
      w = where(sqrt(Xra_bolos_scan(*, boloref(detarr))^2. + Ydec_bolos_scan(*, boloref(detarr))^2.) $
         le cutout(2) - cut_margin, nt)
      if nt ge 1 then begin
         w = where(sqrt(Xra_bolos_scan(*, boloref(detarr))^2. + Ydec_bolos_scan(*, boloref(detarr))^2.) $
            le cutout(2) + cut_margin, nt)
         Xra_bolos_scan = Xra_bolos_scan(w, *)
         Ydec_bolos_scan = Ydec_bolos_scan(w, *)
      endif else begin
         print, 'Error: cutout region not contained within the first scan !'
         print, 'exit'
         flag_exit = 1
         return
      endelse
   endif
   *Xra_bolos(detarr) = Xra_bolos_scan
   *Ydec_bolos(detarr) = Ydec_bolos_scan


   ;;; scan rotation angle with respect to Xra axis:
   if ((not keyword_set(scanrot_deg)) and (not keyword_set(spiral))) then begin
      ntoff = min([100, nt / 20])
      dx = (*Xra_bolos(detarr))(ntoff:nt-1, boloref(detarr)) - (*Xra_bolos(detarr))(0:nt-1-ntoff, boloref(detarr))
      dy = (*Ydec_bolos(detarr))(ntoff:nt-1, boloref(detarr)) - (*Ydec_bolos(detarr))(0:nt-1-ntoff, boloref(detarr))
      vect_scanrot_deg = atan(dy, dx) * 180.D0 / !dpi
      dx = 0
      dy = 0
      if keyword_set(debug) then begin
         window, 0, xs=1200
         otherplot, lindgen(nt-ntoff), vect_scanrot_deg, xtitle='sample index', $
            ytitle='scanning angle wrt X axis (degrees)'
      endif
      med = (min(vect_scanrot_deg) + max(vect_scanrot_deg)) / 2.
      if nt gt 12 * 100 then begin
         if median(vect_scanrot_deg(100:6*100)) gt med then w = where(vect_scanrot_deg gt med) $
            else w = where(vect_scanrot_deg lt med)
      endif else begin
         if median(vect_scanrot_deg(0:nt/2)) gt med then w = where(vect_scanrot_deg gt med) $
            else w = where(vect_scanrot_deg lt med)
      endelse
      histo = histogram(vect_scanrot_deg(w), locations=ang, binsize=1.)
      m = max(histo, ind)
      w = where(abs(vect_scanrot_deg - ang(ind)) le 10.)
      scanrot_deg = (median(vect_scanrot_deg(w)) mod 180.D0)
      list_scanrot(0) = scanrot_deg
      if finite(scanrot_optim) lt 1 then print, 'scan rotation angle wrt EW axis = ' + chain(scanrot_deg, 2) + ' degrees'
      if keyword_set(debug) then begin
         otherplot, /over, color=10, lstyle=2, lindgen(nt-ntoff), scanrot_deg + fltarr(nt-ntoff)
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      endif
      cos_scanrot = cos(scanrot_deg * !dpi / 180.D0)
      sin_scanrot = sin(scanrot_deg * !dpi / 180.D0)

      if finite(scanrot_optim) gt 0 then begin
         p_scanrot_deg = scanrot_optim
         p_cos_scanrot = cos(scanrot_optim * !dpi / 180.D0)
         p_sin_scanrot = sin(scanrot_optim * !dpi / 180.D0)
      endif else begin
         p_scanrot_deg = scanrot_deg
         p_cos_scanrot = cos_scanrot
         p_sin_scanrot = sin_scanrot
      endelse
      p_maprot_deg = 0.D0


      ;;; determine map rotation angle with respect to x axis:
      if (orient eq 'scan' and nblocks eq 1) then maprot_deg = 0.D0 else maprot_deg = scanrot_deg mod 90.D0

      ;;; attempt to eliminate initial slew:
      w_exc = 0
      if nblocks eq 1 then begin
         vect_diff = (abs(vect_scanrot_deg - scanrot_deg) mod 180.)
         w = where(vect_diff le 10. or vect_diff ge 170.)
         if w(0) gt 0 then w_exc = lindgen(w(0))
         w = 0
      endif

      ;;; optionally take rotation in reference header:
      if keyword_set(hdr_ref) then begin
         crota2 = sxpar(hdr_ref, 'CROTA2', count=cc)
         if cc lt 1 then begin
            print, 'missing CROTA2 keyword in reference header: assuming 0.'
            crota2 = 0.D0
         endif
         if abs(crota2) lt 1.e-6 then orient = 'astro' else orient = 'scan'
         maprot_deg = (crota2 + scanrot_deg) mod 90.D0

         scanrot_deg = -crota2
         cos_scanrot = cos(scanrot_deg * !dpi / 180.D0)
         sin_scanrot = sin(scanrot_deg * !dpi / 180.D0)
      endif

;;;      if finite(scanrot_optim) lt 1 then print, 'map rotation angle wrt x axis = ' + chain(maprot_deg, 2) + ' degrees'
   endif

   if keyword_set(spiral) then w_exc = 0
   if keyword_set(w_exc) then begin
      (*Xra_bolos(detarr))(w_exc, *) = !values.f_nan
      (*Ydec_bolos(detarr))(w_exc, *) = !values.f_nan
   endif


   if nblocks gt 1 then if finite(scanrot_optim) gt 0 then begin
      scanrot_deg = scanrot_optim
      cos_scanrot = cos(scanrot_deg * !dpi / 180.D0)
      sin_scanrot = sin(scanrot_deg * !dpi / 180.D0)
   endif


   if ((not keyword_set(scanrot_deg)) and keyword_set(spiral) and keyword_set(hdr_ref)) then begin
      crota2 = sxpar(hdr_ref, 'CROTA2', count=cc)
      if cc lt 1 then begin
         print, 'missing CROTA2 keyword in reference header: assuming 0.'
         crota2 = 0.D0
      endif
      if abs(crota2) lt 1.e-6 then orient = 'astro' else orient = 'scan'
      maprot_deg = 0.D0
      scanrot_deg = -crota2
      cos_scanrot = cos(scanrot_deg * !dpi / 180.D0)
      sin_scanrot = sin(scanrot_deg * !dpi / 180.D0)
   endif


   if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then begin
      *p_Xra_bolos(detarr) = *Xra_bolos(detarr) * p_cos_scanrot + *Ydec_bolos(detarr) * p_sin_scanrot
      *p_Ydec_bolos(detarr) = *Ydec_bolos(detarr) * p_cos_scanrot - *Xra_bolos(detarr) * p_sin_scanrot
      p_off_cen_Xra(detarr, 0) = -min(*p_Xra_bolos(detarr), /nan)
      p_off_cen_Ydec(detarr, 0) = -min(*p_Ydec_bolos(detarr), /nan)
      p_max_xdeg(detarr, 0) = max(*p_Xra_bolos(detarr), /nan)
      p_max_ydeg(detarr, 0) = max(*p_Ydec_bolos(detarr), /nan)
      *p_Xra_bolos(detarr) += p_off_cen_Xra(detarr, 0)
      *p_Ydec_bolos(detarr) += p_off_cen_Ydec(detarr, 0)
   endif

   case orient of
      'astro': begin
         off_cen_Xra(detarr, 0) = -min(*Xra_bolos(detarr), /nan)
         off_cen_Ydec(detarr, 0) = -min(*Ydec_bolos(detarr), /nan)
         end
      'scan': begin
         Xra_tempo = *Xra_bolos(detarr)
         *Xra_bolos(detarr) = *Xra_bolos(detarr) * cos_scanrot + *Ydec_bolos(detarr) * sin_scanrot
         *Ydec_bolos(detarr) = *Ydec_bolos(detarr) * cos_scanrot - Xra_tempo * sin_scanrot
         Xra_tempo = 0
         off_cen_Xra(detarr, 0) = -min(*Xra_bolos(detarr), /nan)
         off_cen_Ydec(detarr, 0) = -min(*Ydec_bolos(detarr), /nan)
         end
   endcase

   drdeg_max(detarr) = sqrt(max((*Xra_bolos(detarr) $
      - rebin((*Xra_bolos(detarr))(*, boloref(detarr)), nt, nbolo(detarr)))^2. $
      + (*Ydec_bolos(detarr) - rebin((*Ydec_bolos(detarr))(*, boloref(detarr)), nt, nbolo(detarr)))^2., /nan))

   max_xdeg(detarr, 0) = max(*Xra_bolos(detarr), /nan)
   max_ydeg(detarr, 0) = max(*Ydec_bolos(detarr), /nan)
endif



if keyword_set(spiral) then warn_vscan = 1

if nscans gt 1 then begin
   w = where(ind_arrays eq 1)
   arr_1 = w(0)
   for kf = 1, nf - 1 do begin
      restore, dir_in + files_scans(kf)
      nt_b = n_elements(scan.time_s)

      if (not keyword_set(warn_vscan)) then if abs(scan.vscan_status / vscan_status - 1.) gt 0.25 then begin
         print, crs + '*******************************************************************************'
         print, crs + 'The input scans have inhomogeneous scan speeds according to the level-1 metadata.'
         print, 'This is not recommended in the general case (but not a major issue).'
         print, crs + '*******************************************************************************'
         warn_vscan = 1
      endif
      if (not keyword_set(warn_dt)) then if abs(scan.dt_status / dt_status - 1.) gt 0.25 then begin
         print, crs + '*******************************************************************************'
         print, crs + 'The input scans have inhomogeneous sampling rates !'
         print, '(according to the level-1 metadata)
         print, 'This is not recommended.'
         print, crs + '*******************************************************************************'
         warn_dt = 1
      endif

      vscan_status = max([vscan_status, scan.vscan_status])
      dt_status = max([dt_status, scan.dt_status])

      for detarr = 0, narr - 1 do if ind_arrays(detarr) eq 1 then begin
         pos = stregex(tags, 'Xra_bolos_' + arrays_default(detarr), /fold_case)
         w = where(pos ge 0)
         Xra_bolos_scan = scan.(w(0))
         pos = stregex(tags, 'Ydec_bolos_' + arrays_default(detarr), /fold_case)
         w = where(pos ge 0)
         Ydec_bolos_scan = scan.(w(0))

         if (not keyword_set(cutout)) then begin
            tab_mid_x(kf) = median(Xra_bolos_scan)
            tab_mid_y(kf) = median(Ydec_bolos_scan)
         endif

         refx = mid_Xra(detarr)
         refy = mid_Ydec(detarr)
         gnomonic, Xra_bolos_scan, Ydec_bolos_scan, refx=refx, refy=refy

         if keyword_set(offset_ra_as) then begin
            Xra_bolos_scan -= (offset_ra_as(kf) / 3600.D0)
            Ydec_bolos_scan += (offset_dec_as(kf) / 3600.D0)
         endif
;;;
   wb = where(total(finite(Xra_bolos_scan), 1) ge 0.9 * max(total(finite(Xra_bolos_scan), 1)) $
      and total(finite(Ydec_bolos_scan), 1) ge 0.9 * max(total(finite(Ydec_bolos_scan), 1)))
   if min(abs(wb - boloref(detarr))) gt 0 then begin
      m = min((Xra_bolos_scan(nt_b/2L, wb) - median(Xra_bolos_scan(nt_b/2L, wb), /even))^2. $
         + (Ydec_bolos_scan(nt_b/2L, wb) - median(Ydec_bolos_scan(nt_b/2L, wb), /even))^2., bref)
      boloref(detarr) = wb(bref)
   endif
;;;
         if keyword_set(cutout) then begin
            w = where(sqrt(Xra_bolos_scan(*, boloref(detarr))^2. + Ydec_bolos_scan(*, boloref(detarr))^2.) $
               le cutout(2) - cut_margin, nt_b)
            if nt_b ge 1 then begin
               w = where(sqrt(Xra_bolos_scan(*, boloref(detarr))^2. + Ydec_bolos_scan(*, boloref(detarr))^2.) $
                  le cutout(2) + cut_margin, nt_b)
               Xra_bolos_scan = Xra_bolos_scan(w, *)
               Ydec_bolos_scan = Ydec_bolos_scan(w, *)
            endif else begin
               print, 'Error: cutout region not contained within scan ' + files_scans(kf) + ' !'
               print, 'exit'
               flag_exit = 1
               return
            endelse
         endif


         ;;; attempt to eliminate initial slew:
         if (detarr eq arr_1 and (not keyword_set(spiral))) then begin
            ntoff = min([100, nt_b / 20])
            dx = Xra_bolos_scan(ntoff:nt_b-1, boloref(detarr)) - Xra_bolos_scan(0:nt_b-1-ntoff, boloref(detarr))
            dy = Ydec_bolos_scan(ntoff:nt_b-1, boloref(detarr)) - Ydec_bolos_scan(0:nt_b-1-ntoff, boloref(detarr))
            vect_scanrot_deg = atan(dy, dx) * 180.D0 / !dpi
            dx = 0
            dy = 0
            med = (min(vect_scanrot_deg) + max(vect_scanrot_deg)) / 2.
            if nt_b gt 12 * 100 then begin
               if median(vect_scanrot_deg(100:6*100)) gt med then w = where(vect_scanrot_deg gt med) $
                  else w = where(vect_scanrot_deg lt med)
            endif else begin
               if median(vect_scanrot_deg(0:nt_b/2)) gt med then w = where(vect_scanrot_deg gt med) $
                  else w = where(vect_scanrot_deg lt med)
            endelse
            histo = histogram(vect_scanrot_deg(w), locations=ang, binsize=1.)
            m = max(histo, ind)
            w = where(abs(vect_scanrot_deg - ang(ind)) le 10.)
            scanrot_deg_cross = (median(vect_scanrot_deg(w)) mod 180.D0)
            list_scanrot(kf) = scanrot_deg_cross

            w_exc_cross = 0
            if nblocks eq 1 then begin
               vect_diff = (abs(vect_scanrot_deg - scanrot_deg_cross) mod 180.)
               w = where(vect_diff le 10. or vect_diff ge 170.)
               if w(0) gt 0 then w_exc_cross = lindgen(w(0))
               w = 0
            endif
         endif

         if keyword_set(spiral) then w_exc_cross = 0
         if keyword_set(w_exc_cross) then begin
            Xra_bolos_scan(w_exc_cross, *) = !values.f_nan
            Ydec_bolos_scan(w_exc_cross, *) = !values.f_nan
         endif


         if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then begin
            p_Xra_bolos_scan = Xra_bolos_scan * p_cos_scanrot + Ydec_bolos_scan * p_sin_scanrot
            p_Ydec_bolos_scan = Ydec_bolos_scan * p_cos_scanrot - Xra_bolos_scan * p_sin_scanrot
            p_off_cen_Xra(detarr, kf) = -min(p_Xra_bolos_scan, /nan)
            p_off_cen_Ydec(detarr, kf) = -min(p_Ydec_bolos_scan, /nan)
            p_max_xdeg(detarr, kf) = max(p_Xra_bolos_scan, /nan)
            p_max_ydeg(detarr, kf) = max(p_Ydec_bolos_scan, /nan)
         endif

         case orient of
            'astro': begin
               off_cen_Xra(detarr, kf) = -min(Xra_bolos_scan, /nan)
               off_cen_Ydec(detarr, kf) = -min(Ydec_bolos_scan, /nan)
               end
           'scan': begin
               Xra_tempo = Xra_bolos_scan
               Xra_bolos_scan = Xra_bolos_scan * cos_scanrot + Ydec_bolos_scan * sin_scanrot
               Ydec_bolos_scan = Ydec_bolos_scan * cos_scanrot - Xra_tempo * sin_scanrot
               Xra_tempo = 0
               off_cen_Xra(detarr, kf) = -min(Xra_bolos_scan, /nan)
               off_cen_Ydec(detarr, kf) = -min(Ydec_bolos_scan, /nan)
               end
         endcase

         drdeg_max(detarr) = max([drdeg_max(detarr), $
            sqrt(max((Xra_bolos_scan - rebin(Xra_bolos_scan(*, boloref(detarr)), nt_b, nbolo(detarr)))^2. $
            + (Ydec_bolos_scan - rebin(Ydec_bolos_scan(*, boloref(detarr)), nt_b, nbolo(detarr)))^2., /nan))])

         max_xdeg(detarr, kf) = max(Xra_bolos_scan, /nan)
         max_ydeg(detarr, kf) = max(Ydec_bolos_scan, /nan)
      endif
   endfor

   Xra_bolos_scan = 0
   Ydec_bolos_scan = 0
endif



;;; re-order scans to have alternating angles:
;;if (finite(scanrot_optim) lt 1 and nscans gt 2 and (not keyword_set(spiral))) then begin
;;   ind_order_scans = order_angles(list_scanrot)
;;   print, crs + 'scan ordering:'
;;   if max(abs(ind_order_scans - indgen(nscans))) gt 0 then $
;;      for s = 0, nscans - 1 do print, files_scans(ind_order_scans(s)) $
;;      else print, 'unchanged'
;;   print, crs
;;endif

ind_order_scans = indgen(nscans)



;;;
if finite(scanrot_optim) lt 1 then begin
   w = where(ind_arrays eq 1)
   w = w(0)
   med_x = (max_xdeg(w, *) - off_cen_Xra(w, *)) / 2.
   med_y = (max_ydeg(w, *) - off_cen_Ydec(w, *)) / 2.
   dev_x = sqrt(mean((med_x - min(med_x))^2.))
   dev_y = sqrt(mean((med_y - min(med_y))^2.))
   size_x = median(max_xdeg(w, *) + off_cen_Xra(w, *), /even)
   size_y = median(max_ydeg(w, *) + off_cen_Ydec(w, *), /even)
   if keyword_set(visu) then begin
      print, ''
      print, '     dev_x / size_x = ' + chain(dev_x / size_x, 3)
      print, '     dev_y / size_y = ' + chain(dev_y / size_y, 3)
   endif
   if (dev_x gt size_x or dev_y gt size_y) then begin
      scanrot_optim = 0.D0
      if nblocks gt 1 then begin
         alpha = atan(tab_mid_y - median(tab_mid_y, /even), tab_mid_x - median(tab_mid_x, /even)) * 180.D0 / !dpi
         w = where(alpha lt 0., cw)
         if cw ge 1 then alpha(w) += 180.D0
         scanrot_optim = median(alpha(where(abs(tab_mid_y) gt 0. and abs(tab_mid_x) gt 0.)), /even)
      endif
      restore, dir_in + files_scans(0)
      goto, re_read_scans
   endif
endif
;;;

;;; deal with non-standard observations with non-orthogonal orientations:
if (finite(scanrot_optim) lt 1 and nscans gt 1 and (not keyword_set(spiral))) then begin
   n_or = 2
   for kf = 1, nf - 1 do begin
      tab_diff = abs(list_scanrot(kf) - list_scanrot(0:kf))
      tab_diff = min([[tab_diff mod 90.D0], [abs(90.D0 - tab_diff) mod 90.D0]], dimension=2)
      w = where(tab_diff gt 5.D0, cw)
      if cw ge kf then n_or += 1
   endfor
   if n_or ge 3 then begin
      print, '     deriving the optimal processing orientation'
      n_r = 180
      min_scanrot = min(list_scanrot)
      tab_rot = (min_scanrot + dindgen(n_r) / 2.D0) ## (1.D0 + dblarr(nf))
      tab_diff = abs(tab_rot - rebin(list_scanrot, nf, n_r))
      tab_diff = min([[[tab_diff mod 90.D0]], [[abs(90.D0 - tab_diff) mod 90.D0]]], dimension=3)
      tab_devmax = max(tab_diff, dimension=1)
      m = min(tab_devmax, ind)
      scanrot_optim = min_scanrot + double(ind) / 2.D0
      if (abs(scanrot_optim - scanrot_deg) mod 90.D0) gt 5.D0 then begin
         restore, dir_in + files_scans(0)
         goto, re_read_scans
      endif else scanrot_optim = scanrot_deg
   endif
endif else if finite(scanrot_optim) gt 0 then $
   print, '     optimal angle = ' + chain(scanrot_optim, 2) + ' degrees'


if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then ptr_free, p_Xra_bolos, p_Ydec_bolos
;;;
ptr_free, Xra_bolos, Ydec_bolos
;;;



;;; take into account possible spatial shifts between reference scans:

if nf gt 1 then begin
   off_cen_Xra = max(off_cen_Xra, dimension=2)
   off_cen_Ydec = max(off_cen_Ydec, dimension=2)
   max_xdeg = max(max_xdeg, dimension=2)
   max_ydeg = max(max_ydeg, dimension=2)
endif
max_xdeg += off_cen_Xra
max_ydeg += off_cen_Ydec
if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then begin
   if nf gt 1 then begin
      p_off_cen_Xra = max(p_off_cen_Xra, dimension=2)
      p_off_cen_Ydec = max(p_off_cen_Ydec, dimension=2)
      p_max_xdeg = max(p_max_xdeg, dimension=2)
      p_max_ydeg = max(p_max_ydeg, dimension=2)
   endif
   p_max_xdeg += p_off_cen_Xra
   p_max_ydeg += p_off_cen_Ydec
endif


;;; reference = middle of larger map:

if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then begin
   w = where(ind_arrays eq 1)
   m = max(p_max_xdeg(w)^2. + p_max_ydeg(w)^2., ind)
   w = w(ind)
   p_mid_Xra = mid_Xra(w)
   p_mid_Ydec = mid_Ydec(w)
   p_off_cen_Xra = p_off_cen_Xra(w)
   p_off_cen_Ydec = p_off_cen_Ydec(w)
endif

if keyword_set(hdr_ref) then begin
   crval1 = sxpar(hdr_ref, 'CRVAL1')
   crval2 = sxpar(hdr_ref, 'CRVAL2')
   crpix1 = sxpar(hdr_ref, 'CRPIX1') - 0.5
   crpix2 = sxpar(hdr_ref, 'CRPIX2') - 0.5
   cdelt1 = sxpar(hdr_ref, 'CDELT1', count=cc)
   if cc gt 0 then pix_deg = sqrt(abs(cdelt1) * sxpar(hdr_ref, 'CDELT2')) $
      else pix_deg = sqrt((sxpar(hdr_ref, 'CD1_1'))^2. + (sxpar(hdr_ref, 'CD1_2'))^2.)
   mid_Xra = -crval1
   mid_Ydec = crval2
   off_cen_Xra = crpix1 * pix_deg
   off_cen_Ydec = crpix2 * pix_deg
endif else begin
   w = where(ind_arrays eq 1)
   m = max(max_xdeg(w)^2. + max_ydeg(w)^2., ind)
   w = w(ind)
   mid_Xra = mid_Xra(w)
   mid_Ydec = mid_Ydec(w)
   off_cen_Xra = off_cen_Xra(w)
   off_cen_Ydec = off_cen_Ydec(w)
endelse



if (not keyword_set(hdr_ref)) then begin
   crval1 = -mid_Xra
   crval2 = mid_Ydec
endif



;;; prepare to slice data into spatial blocks:

w = where(ind_arrays_eff eq 1)
w = w(0)
margin = 1. * drdeg_max(w)
if ((ntb gt max_n_samples or nblocks gt 1) and (not keyword_set(cutout))) then begin
   if nblocks le 1 then nblocks = 1L + ntb / max_n_samples else nblocks = long(nblocks)
if (orient eq 'astro' or keyword_set(hdr_ref)) then begin
   dxdeg = p_max_xdeg(w)
   dydeg = p_max_ydeg(w)
endif else begin
   dxdeg = max_xdeg(w)
   dydeg = max_ydeg(w)
endelse
   min_Xra_scan = drdeg_max(w)
   min_Ydec_scan = drdeg_max(w)

   dxydeg_block = sqrt(dxdeg * dydeg / double(nblocks))
   if ((dxdeg le 1.2 * dxydeg_block or dev_y gt size_y) and dxdeg le dydeg) then begin
      dxdeg_block = dxdeg
      dydeg_block = dydeg / double(nblocks)
      nxbl = 1
      nybl = nblocks
   endif else if ((dydeg le 1.2 * dxydeg_block or dev_x gt size_x) and dydeg lt dxdeg) then begin
      dydeg_block = dydeg
      dxdeg_block = dxdeg / double(nblocks)
      nybl = 1
      nxbl = nblocks
   endif else if dxdeg le dydeg then begin
      nxbl = floor(dxdeg / dxydeg_block) + 1
      dxdeg_block = dxdeg / double(nxbl)
      nybl = ceil(nblocks / float(nxbl))
      dydeg_block = dydeg / double(nybl)
      nblocks = nxbl * nybl
   endif else begin
      nybl = floor(dydeg / dxydeg_block) + 1
      dydeg_block = dydeg / double(nybl)
      nxbl = ceil(nblocks / float(nybl))
      dxdeg_block = dxdeg / double(nxbl)
      nblocks = nxbl * nybl
   endelse

   xdeg_blocks = dblarr(nblocks, 2)
   ydeg_blocks = dblarr(nblocks, 2)
   tab_nxbl = lindgen(nxbl)
   for ybl = 0, nybl - 1 do begin
      xdeg_blocks(tab_nxbl+nxbl*ybl, 0) = min_Xra_scan + dxdeg_block * tab_nxbl - margin
      ydeg_blocks(tab_nxbl+nxbl*ybl, 0) = min_Ydec_scan + dydeg_block * ybl - margin
   endfor
   xdeg_blocks(*, 1) = xdeg_blocks(*, 0) + dxdeg_block + 2.D0 * margin
   ydeg_blocks(*, 1) = ydeg_blocks(*, 0) + dydeg_block + 2.D0 * margin

   tab_nybl = lindgen(nybl)
   xdeg_blocks(nxbl*tab_nybl, 0) -= drdeg_max(w)
   xdeg_blocks(nxbl-1+nxbl*tab_nybl, 1) += drdeg_max(w)
   ydeg_blocks(tab_nxbl, 0) -= drdeg_max(w)
   ydeg_blocks(tab_nxbl+nxbl*(nybl-1), 1) += drdeg_max(w)
   tab_nxbl = 0
   tab_nybl = 0

endif else nblocks = 1



files_scans = files_scans(ind_order_scans)
list_scanrot = list_scanrot(ind_order_scans)
if keyword_set(offset_ra_as) then begin
   offset_ra_as = offset_ra_as(ind_order_scans)
   offset_dec_as = offset_dec_as(ind_order_scans)
endif


if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral))) then save, file=file_param_blocks, $
   p_scanrot_deg, p_maprot_deg, p_off_cen_Xra, p_off_cen_Ydec, p_mid_Xra, p_mid_Ydec, p_max_xdeg, p_max_ydeg, $
   scanrot_deg, maprot_deg, off_cen_Xra, off_cen_Ydec, mid_Xra, mid_Ydec, drdeg_max, max_xdeg, max_ydeg, $
   crval1, crval2, /xdr $
 else save, file=file_param_blocks, scanrot_deg, maprot_deg, $
    off_cen_Xra, off_cen_Ydec, mid_Xra, mid_Ydec, drdeg_max, max_xdeg, max_ydeg, crval1, crval2, /xdr


end
