pro proj_map_artemis, map_signal, map_weight, map_error, file_scalars=file_scalars, nstep_imprint_final=nstep_imprint_final, $
              file_imprint_final=file_imprint_final, dimx_proj=dimx_proj, dimy_proj=dimy_proj, $
              maxnoise=maxnoise, ind_scans=ind_scans, comp_errors=comp_errors, lowfnoise=lowfnoise, $
              todisk=todisk, cut_nza_weight=cut_nza_weight, ind_subscans=ind_subscans, cube_mask=cube_mask


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars
ntb = nt * nb

if keyword_set(nstep_imprint_final) then nstep_imprint = nstep_imprint_final
if keyword_set(file_imprint_final) then file_imprint = file_imprint_final


if not keyword_set(dimx_proj) then begin
   dimx_proj = dimx
   dimy_proj = dimy
endif
dimxy = dimx_proj * dimy_proj

map_weight = fltarr(dimx_proj, dimy_proj)
map_signal = fltarr(dimx_proj, dimy_proj)
if keyword_set(comp_errors) then begin
   map_error = fltarr(dimx_proj, dimy_proj)
   map_sqw = fltarr(dimx_proj, dimy_proj)
   map_lowfnoise = fltarr(dimx_proj, dimy_proj)
   map_lfn_2comp = fltarr(dimx_proj, dimy_proj, 2)
endif else begin
   map_error = 0
   map_lowfnoise = 0
   map_lfn_2comp = 0
endelse
if keyword_set(cube_mask) then map_mask = fltarr(dimx_proj, dimy_proj) else map_mask = 0

weight = 1. / maxnoise^2.
w = where(finite(weight) gt 0)
weight /= mean(weight(w))


tab_subscan = lonarr(ind_subscans(1))
for i = 1L, nsubscans - 1L do tab_subscan = [temporary(tab_subscan), i + lonarr(ind_subscans(i+1) - ind_subscans(i))]


restore, file_pixcoords

indimprint = nstep_imprint * floor(nstep_imprint * (y_bolos - floor(y_bolos))) $
   + floor(nstep_imprint * (x_bolos - floor(x_bolos)))
x_bolos = 0
y_bolos = 0


restore, file_projfine

restore, file_imprint

sq_dim_imprint = dim_imprint^2
imprint_reform = reform(imprint, nstep_imprint^2, sq_dim_imprint)
imprint = 0


restore, file_signal
if (keyword_set(comp_errors) and (not keyword_set(lowfnoise))) then lowfnoise = fltarr(nt, nb, 3)


for k = 0L, nproj_fine - 1L do begin

   ;;; quicker way of evaluating: wp = where(indxy_beam eq wproj_fine(k), cp):
   revind_0 = revind_fine(wproj_fine(k))
   revind_1 = revind_fine(wproj_fine(k)+1)
   cp = revind_1 - revind_0
   wp = revind_fine(revind_0:revind_1-1)

   vectpos = vectpos_startimprint(k) + vectpos_imprint
   bin_weight = rebin(weight(nsubscans * (wp / nt) + tab_subscan(wp mod nt)), cp, sq_dim_imprint) $
      * imprint_reform(indimprint(wp), *)
   bin_signal = rebin(signal(wp), cp, sq_dim_imprint)
   map_weight(vectpos) += total(bin_weight, 1)
   map_signal(vectpos) += total(bin_signal * bin_weight, 1)

   if keyword_set(comp_errors) then begin
      map_sqw(vectpos) += total(bin_weight^2., 1)
      map_error(vectpos) += total(bin_signal^2. * bin_weight, 1)
      map_lowfnoise(vectpos) += total(rebin(lowfnoise(wp), cp, sq_dim_imprint) * bin_weight, 1)
      map_lfn_2comp(vectpos) += total(rebin(lowfnoise(ntb+wp), cp, sq_dim_imprint) * bin_weight, 1)
      map_lfn_2comp(dimxy+vectpos) += total(rebin(lowfnoise(2L*ntb+wp), cp, sq_dim_imprint) * bin_weight, 1)
   endif

   if keyword_set(cube_mask) then map_mask(wproj_fine(k)) = total(cube_mask(wp)) / float(cp)
endfor


wcov = where(map_weight gt 0.)
if keyword_set(comp_errors) then begin
   map_error(wcov) = (map_error(wcov) * map_weight(wcov) - map_signal(wcov)^2.) $
      / (map_weight(wcov)^2. - map_sqw(wcov))
   map_lowfnoise(wcov) /= map_weight(wcov)
   map_lfn_2comp(wcov) /= map_weight(wcov)
   map_lfn_2comp(dimxy+wcov) /= map_weight(wcov)
endif
map_signal(wcov) /= map_weight(wcov)

if keyword_set(comp_errors) then map_error(wcov) = sqrt(map_error(wcov) / map_weight(wcov))


if (n_params() eq 0 or keyword_set(todisk)) then $
   save, file=file_maps, map_signal, map_weight, map_error, map_lowfnoise, map_lfn_2comp, map_mask, /xdr

end
