pro scanam_artemis, dir_scanlist=dir_scanlist, visu=visu, vis_traject=vis_traject, debug=debug, $
                    galactic=galactic, offset_ra_as=offset_ra_as, offset_dec_as=offset_dec_as, $
                    hdr_ref=hdr_ref, cutout=cutout, nblocks=nblocks, block_start=block_start, $
                    one_plane_fits=one_plane_fits, $
;;;
;;;                    spiral=spiral, $   ;;; not functional yet
;;;***;;;
;;; keywords used in batch mode only:
                  batch=batch, file_out=file_out, orient=orient, pixsize=pixsize, nzadata=nzadata, array=array, $

;;;***;;;
;;; keywords for local use only (for technical tests and simulations):
                  noglitch=noglitch, nodrifts=nodrifts, nocorr=nocorr, save_final=save_final


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation of the code:
;;; http://www2.iap.fr/users/roussel/artemis/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; first public delivery: March 2018

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; to be used with input structures formatted from level 1 files
;;; by  convert_input_subscans_artemis.pro
;;;
;;; directory and file names of the input structures in the ascii file  scanlist_artemis
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; automatically slices the data into overlapping spatial blocks when the data volume has to be cut;
;;; maximum block volume depends on the "max_n_samples" parameter
;;; (function of machine characteristics, has to be calibrated)
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



;;;;;;;;;
;;;cpu, tpool_nthreads=1
;cpu, tpool_min_elts=long64(1000000L), tpool_nthreads=8L
;;;;;;;;;

resolve_all, /continue_on_error, /quiet
if (keyword_set(visu) or keyword_set(debug)) then loadct, 15, /silent

crs = string(13b) + string(10b)


version = 'v3.1 -- May 2018'
vers = strtrim(strmid(version, 0, 5), 2)
print, ''
print, 'running Scanamorphos for Artemis ' + version
print, ''



;;;;;;;;;;;;;;;;
;;; definitions
;;;;;;;;;;;;;;;;


inst = 'artemis'
narr = 2
arrays_names = ['350', '450']
arrays_label = ['350 microns', '450 microns']
nbolo = [64L*36L, 64L*36L]
fwhm_as = [8., 10.]
pfov = [3.8, 3.8]   ;;; --> superseded by metadata of the first scan
p_arrhw = [1., 1.]
;;;
;;;boloref = [31L+64L*17L, 31L+64L*17L]
boloref = [24L+64L*8L, 24L+64L*8L]   ;;; superseded in prepare_blocks_artemis according to valid detectors
;;;
fact_scales = [2., 2.]
;;fact_scales = [4., 4.]
;;;
tab_npass_crit = [3, 3]
fact_npass_cen = 0.72
;;;
logf_inf = 1.
logf_sup = 1.3
logf_high_inf = 0.7
logf_high_sup = 1.
;;;


iter_aver_max = 1

fact_fwhm = [0.1, 0.1]   ;;; fact_fwhm * FWHM = stability length (automatically increased if needed)
min_nsamples = [7, 7]   ;;; minimum number of samples per stability length crossing
;;;
if keyword_set(spiral) then min_nsamples = [8, 8]
;;;
tc_drift_s = 0.05          ;;; minimum timescale for drifts (automatically increased as needed)



if (not keyword_set(dir_scanlist)) then dir_scanlist = './'
nc = strlen(dir_scanlist)
if strmid(dir_scanlist, nc-1) ne '/' then dir_scanlist = dir_scanlist + '/'
dir_grids = dir_scanlist

file_scanlist = 'scanlist_' + inst
test = file_test(dir_scanlist + file_scanlist, /read)
if test eq 0 then begin
   print, crs + 'file ' + file_scanlist + ' inexistent or non-readable: exit'
   goto, end_label
endif

if (not keyword_set(array)) then band = 350 else band = long(array)
arrays_default = ['PSW', 'PLW']

ind_arrays = intarr(narr)
for detarr = 0, narr - 1 do begin
   w = where(chain(band) eq arrays_names(detarr), cw)
   if cw ge 1 then ind_arrays(detarr) = 1
endfor



;;; following parameter depends on machine characteristics

;;; if nt * nb > max_n_samples, then the field of view will be sliced into several blocks:
;;;max_n_samples = long(1.3D+9)   ;;; maximum tested with MemTotal = 48 Gb
if !version.memory_bits ge 64 then max_n_samples = long64(5.8D+9) $
   else max_n_samples = long(1.3D+9)
;;; maximum tested so far (with MemTotal = 265 Gb)


;;; if number of coarse time steps > max_nt_coarse,
;;; then the minimum timescale of the average drift will be increased:
;;;   square of long(4.6D+4) ~= long integer limit
;;;   Only with a 64-bit version can max_nt_coarse be increased.
;if !version.memory_bits ge 64 then max_nt_coarse = long64(2.5D+5) else max_nt_coarse = long(4.6D+4)
;;;
max_nt_coarse = long64(1.D+5)
;;; Use e.g. max_nt_coarse = long64(1.D+5) to reduce memory usage for deep observations
;;;


if not keyword_set(debug) then debug = 0
if not keyword_set(visu) then visu = 0
if keyword_set(debug) then visu = 1

if (not keyword_set(nblocks)) then nblocks = 1



;;;;;;;;;;;;;;;;;;;;
;;; user parameters
;;;;;;;;;;;;;;;;;;;;


dir_in = ''
files_scans = ''
file = ''
openr, 1, dir_scanlist + file_scanlist
readf, 1, format='(A)', dir_in
dir_in = strtrim(dir_in, 2)
nc = strlen(dir_in)
if strmid(dir_in, nc-1) ne '/' then dir_in = dir_in + '/'
readf, 1, format='(A)', files_scans
files_scans = strcompress(files_scans, /remove_all)
while (not eof(1)) do begin
   readf, 1, format='(A)', file
   files_scans = [files_scans, strcompress(file, /remove_all)]
endwhile
close, 1
nscans = n_elements(files_scans)



;;; select the detector array corresponding to the data:
restore, dir_in + files_scans(0)
tags = tag_names(scan)
for detarr = 0, narr - 1 do begin
   pos = stregex(tags, 'signal_' + arrays_default(detarr), /fold_case)
   w = where(pos ge 0, cw)
   if cw ge 1 then if (ind_arrays(detarr) lt 1 or total(ind_arrays) gt 1) then begin
      print, crs + '****************************************************'
      print, crs + 'The selected or default detector array does not match the data:
      print, crs + 'Selecting ' + arrays_label(detarr) + ' instead.'
      print, crs + '****************************************************'
      print, crs
      ind_arrays = intarr(narr)
      ind_arrays(detarr) = 1
      band = long(arrays_names(detarr))
   endif
   if cw ge 1 then begin
      array = [arrays_default(detarr)]
      pfov(detarr) = scan.pfov
      target = scan.target
      obsproject = scan.obsproject
   endif
endfor
scan = 0



scannum = long(stregex(files_scans(0), '[0-9]{5,10}', /extract))
get_calib_project_info, scannum, init_obs_str, band=band, /silent

dir_out_processed = init_obs_str.apexdata
nc = strlen(dir_out_processed)
if strmid(dir_out_processed, nc-1) ne '/' then dir_out_processed = dir_out_processed + '/'
dir_out_processed = dir_out_processed + 'map_otf_xdr/' + obsproject + '/' + strtrim(string(band), 2) + '/'
res = file_search(dir_out_processed, /fold_case, /test_write, count=ndir)
if ndir ne 1 then begin
   dir_out_processed = 0
   print, ''
   print, '!!! WARNING !!!'
   print, 'Files in the directory containing the subscan structures cannot be modified.'
   print, 'The processed time series will not be saved.'
   print, ''
endif



if (keyword_set(offset_ra_as) or keyword_set(offset_dec_as)) then begin
   if (n_elements(offset_ra_as) ne nscans or n_elements(offset_dec_as) ne nscans) then begin
      print, 'The number of astrometric offsets does not match the number of scans: exit.'
      goto, end_label
   endif
endif

if keyword_set(cutout) then begin
   if n_elements(cutout) ne 3 then begin
      print, 'The number of elements of "cutout" must be 3:'
      print, '   RA of center, DEC of center, minimum radius of subfield to be mapped'
      print, '   (all in degrees)'
      goto, end_label
   endif
   hdr_ref = 0
   nblocks = 1
   cutout(2) = max([cutout(2), 0.1])
endif


dir_out_default = dir_scanlist
file_out_default = target

orient_default = 'astro'
pix_as_default = fwhm_as / 4.
nzadata_default = 'yes'



for i = 0, nscans - 1 do begin
   test = file_test(dir_in + files_scans(i), /read)
   if test eq 0 then begin
      print, crs + 'input file ' + chain(i+1) + ' inexistent or non-readable: exit'
      goto, end_label
   endif
endfor
print, crs + chain(nscans) + ' scan(s) to be combined'



if keyword_set(batch) then begin
   dir_out = dir_in
;   if (not keyword_set(array)) then array = arrays_default(0) else begin
;      w = where(arrays_names eq chain(band), cw)
;      if cw eq 1 then array = arrays_default(w) else array = arrays_default(0)
;   endelse
   if (not keyword_set(orient)) then orient = orient_default
   if keyword_set(pixsize) then begin
      if n_elements(pixsize) eq 1 then pixsize = pixsize(0) + fltarr(narr)
      tab_pix_as = pixsize
   endif else tab_pix_as = pix_as_default
   if (not keyword_set(nzadata)) then nzadata = nzadata_default
endif else begin

   dir_out = ''
   read, dir_out, prompt = crs + 'output directory:' + crs + '   default = ' + dir_out_default + ' [] ;' $
      + crs + '      choice = '
   if not keyword_set(dir_out) then dir_out = dir_out_default $
      else if strmid(dir_out, strlen(dir_out) - 1, 1) ne '/' then dir_out = dir_out + '/'

   file_out = ''
   read, file_out, prompt = crs + 'output FITS file root:' + crs + '   default = ' + file_out_default + ' [] ;' $
      + crs + '      choice = '
   if not keyword_set(file_out) then file_out = file_out_default
endelse

test = file_test(dir_out, /directory, /write)
if test eq 0 then begin
   print, crs + 'output directory non-existent or non-writable: exit'
   goto, end_label
endif


if (not keyword_set(batch)) then begin
   if keyword_set(spiral) then orient = orient_default else begin
      orient = ''
      if (not keyword_set(hdr_ref)) then read, orient, prompt = crs + 'map orientation:' + crs + $
         'either along the scan or in the standard astronomical frame (N up, E left)' $
         + crs + '   default = ' + orient_default + ' [] ;' $
         + crs + '      choice = [ /scan/astro] '
      if not keyword_set(orient) then orient = orient_default $
         else if strlowcase(strmid(orient, 0, 1)) eq 's' then orient = 'scan' $
         else orient = 'astro'
   endelse

   choice = ''
   if (not keyword_set(hdr_ref)) then read, choice, prompt = crs + 'pixel size (arcsec):' $
      + crs + '   default (FWHM/4) at 350 ; 450 microns = ' + strjoin(chain(pix_as_default, 2), ' ; ') + ' [] ;' $
      + crs + '      choice = [ /new] '
   if not keyword_set(choice) then tab_pix_as = pix_as_default else begin
      tab_pix_as = pix_as_default
      pas = 0.
      for i = 0, narr - 1 do begin
         w = where(array eq arrays_default(i), cw)
         if cw ge 1 then begin
            read, pas, prompt = 'for ' + arrays_label(i) + ': '
            tab_pix_as(i) = float(pas)
         endif
      endfor
   endelse

   if keyword_set(spiral) then nzadata = nzadata_default else begin
      nzadata = ''
      read, nzadata, prompt = crs + 'include data taken while acceleration is non-zero ? ' + $
         crs + '(note: always used for drifts determination,' + crs + 'but can be rejected before final projection)' $
         + crs + '   default = ' + nzadata_default + ' [] ;' $
         + crs + '      choice = [ /yes/no] '
      if not keyword_set(nzadata) then nzadata = nzadata_default $
         else if strlowcase(strmid(nzadata, 0, 1)) eq 'y' then nzadata = 'yes' $
         else nzadata = 'no'
   endelse
   print, crs + crs
endif



dir_tempo = dir_in
file_scalars = dir_tempo + 'save_scalars.xdr'
file_times = dir_tempo + 'save_time_vectors.xdr'
file_time_coarse = dir_tempo + 'save_time_coarse.xdr'
file_projcoarse = dir_tempo + 'save_projcoarse.xdr'
file_projfine = dir_tempo + 'save_projfine.xdr'
file_pixcoords = dir_tempo + 'save_pixcoords.xdr'
file_vscan = dir_tempo + 'save_vscan.xdr'
file_ntmin_pass = dir_tempo + 'save_ntmin_pass.xdr'
file_maps = dir_tempo + 'save_maps.xdr'
file_XYscan = dir_tempo + 'save_XYscan.xdr'

root_signal = dir_tempo + 'save_signal_'
root_xyas = dir_tempo + 'save_xyas_'



;;; slightly different criterion to reject non-uniform emission for shallow observations:
if nscans lt 4 then fact_npass_cen -= 0.02


;;; pixel grids used for the processing (different from pixel grid for final projection):
;;;   coarse grid to compute drift differences: FWHM
;;;   finer grid to project intermediate maps: FWHM / fact_scales
;;; Both scales are multiplied by fact_fwhm: can be as low as 0.1,
;;;   can be larger than 0.5 to improve statistics within stability length for shallow observations
;;;  (high scan speed with low sampling rate).


;;; pixel size for final projection:
tab_pix_as_final = tab_pix_as


;;; diameter of the bolometer imprint on the sky map (assumed to be a uniform disk):
tab_diam_as_imprint = fwhm_as / 2. / sqrt(!pi)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; initial data readout to determine coordinates frame
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


file_param_blocks = dir_tempo + 'save_param_blocks.xdr'
flag_exit = 0

if keyword_set(hdr_ref) then if n_elements(hdr_ref) eq 1 then begin
   test = file_test(dir_in + hdr_ref, /read)
   if test eq 0 then begin
      print, crs + 'reference header inexistent or non-readable: exit'
      goto, end_label
   endif else restore, dir_in + hdr_ref
endif

prepare_blocks_artemis,  file_param_blocks, ind_arrays_eff, orient, nt, nblocks, xdeg_blocks, ydeg_blocks, $
   narr=narr, nscans=nscans, nbolo=nbolo, boloref=boloref, arrays_default=arrays_default, $
   max_n_samples=max_n_samples, ind_arrays=ind_arrays, debug=debug, $
   dir_in=dir_in, files_scans=files_scans, hdr_ref=hdr_ref, cutout=cutout, $
   file_XYscan=file_XYscan, scanrot_optim=scanrot_optim, vscan_status, dt_status, $
   flag_exit=flag_exit, list_scanrot=list_scanrot, offset_ra_as=offset_ra_as, offset_dec_as=offset_dec_as, $
   spiral=spiral

if keyword_set(flag_exit) then goto, end_label

restore, file_param_blocks


;;; pre-determine the stability length based on information in level-1 data:

if (keyword_set(vscan_status) and keyword_set(dt_status)) then begin
   fact_fwhm = ceil(10. * min_nsamples * vscan_status * dt_status / fwhm_as) / 10.
   w = where(fact_fwhm ge 1. and ind_arrays_eff gt 0, cw)
   if cw ge 1 then fact_scales(w) = 2. * fact_fwhm(w)
   print, crs + 'Value of the stability length:'
   for k = 0, narr - 1 do if ind_arrays_eff(k) gt 0 then $
      print, '     ' + chain(fact_fwhm(k), 1) + ' FWHM for ' + arrays_label(k) + crs
endif


;;; optionally come back here to increase stability length
;;; and improve sample statistics:

fact_fwhm_label:


tab_pix_as = tab_pix_as_final

;;; fine grid pixel size:
w = where(abs(1. - fact_scales * tab_pix_as / (fact_fwhm * fwhm_as)) gt 0.1, cw)
if cw ge 1 then tab_pix_as(w) = fact_fwhm(w) * fwhm_as(w) / fact_scales(w)

;;; time gap between successive scans (crossing time of at least 10 FWHM(PLW) for v_scan >= 10"/s):
timegap = fact_fwhm(narr-1) * fwhm_as(narr-1)



w = where(ind_arrays eq 1)
ind = max(w)

off_map_as = 4. * fwhm_as(ind)

dimxas = float(max_xdeg(ind) * 3600.D0) + 2. * off_map_as
dimyas = float(max_ydeg(ind) * 3600.D0) + 2. * off_map_as
if keyword_set(hdr_ref) then begin
   dimx_final = sxpar(hdr_ref, 'NAXIS1')
   dimy_final = sxpar(hdr_ref, 'NAXIS2')
   crpix1_final = sxpar(hdr_ref, 'CRPIX1') - 0.5
   crpix2_final = sxpar(hdr_ref, 'CRPIX2') - 0.5
endif else begin
   dimx_final = ceil(dimxas / tab_pix_as_final(ind))
   dimy_final = ceil(dimyas / tab_pix_as_final(ind))
   crpix1_final = (float(off_cen_Xra * 3600.D0) + off_map_as) / tab_pix_as_final(ind)
   crpix2_final = (float(off_cen_Ydec * 3600.D0) + off_map_as) / tab_pix_as_final(ind)
endelse


;;; common grid for multiple blocks:

if nblocks gt 1 then begin
   dimxas = float(max_xdeg(ind) * 3600.D0) + 2. * off_map_as
   dimyas = float(max_ydeg(ind) * 3600.D0) + 2. * off_map_as
   dimx = ceil(dimxas / tab_pix_as(ind))
   dimy = ceil(dimyas / tab_pix_as(ind))
   crpix1 = (float(off_cen_Xra * 3600.D0) + off_map_as) / tab_pix_as(ind)
   crpix2 = (float(off_cen_Ydec * 3600.D0) + off_map_as) / tab_pix_as(ind)
endif



;;;;;;;;;;;;;;;;;;;
;;; data block loop
;;;;;;;;;;;;;;;;;;;


nscans_tot = nscans
files_scans_tot = files_scans
list_scanrot_tot = list_scanrot


if (not keyword_set(block_start)) then block_start = 0L else block_start = long(block_start)
for block = block_start, nblocks - 1L do begin

   nscans = nscans_tot
   files_scans = files_scans_tot
   list_scanrot = list_scanrot_tot


;;; initialization of the maximum time gap within a subscan:
maxtimediff = 5.

ingest_data_artemis, file_param_blocks, ind_scans, ind_subscans, nsubscans, vscan_as_s, timestep, $
             narr=narr, nscans=nscans, nt=nt, nblocks=nblocks, block=block, $
             inst=inst, nbolo=nbolo, boloref=boloref, ind_arrays_eff=ind_arrays_eff, $
             arrays_default=arrays_default, xdeg_blocks=xdeg_blocks, ydeg_blocks=ydeg_blocks, $
             orient=orient, timegap=timegap, dir_in=dir_in, files_scans=files_scans, $
             hdr_ref=hdr_ref, cutout=cutout, file_XYscan=file_XYscan, file_times=file_times, file_vscan=file_vscan, $
             root_signal=root_signal, root_xyas=root_xyas, maxtimediff=maxtimediff, $
             visu=visu, debug=debug, vis_traject=vis_traject, offset_ra_as=offset_ra_as, offset_dec_as=offset_dec_as, $
             vs_min=vs_min, vs_max=vs_max, spiral=spiral, nzadata=nzadata, elec_boxes=elec_boxes

w = where(strlen(files_scans) gt 0)
files_scans = files_scans(w)
list_scanrot = list_scanrot(w)
elec_boxes = elec_boxes(w, *)



;;;;;;;;;;;;;;;
;;; array loop
;;;;;;;;;;;;;;;


for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
   if nblocks eq 1 then print, crs + crs + 'building ' + arrays_label(detarr) + ' array map'


   ;;;initialize processing time for array:
   t_proc = 0.D0
   time_1 = systime(/seconds)


   if block eq block_start then begin
      nb = nbolo(detarr)
      bref = boloref(detarr)
      if !version.memory_bits ge 64 then begin
         nb = long64(nb)
         bref = long64(bref)
      endif

      pix_as = tab_pix_as(detarr)
      if keyword_set(hdr_ref) then begin
         cdelt1 = sxpar(hdr_ref, 'CDELT1', count=cc)
         if cc gt 0 then tab_pix_as_final(detarr) = float(3600. * sqrt(abs(cdelt1) * sxpar(hdr_ref, 'CDELT2'))) $
            else tab_pix_as_final(detarr) = float(3600. * sqrt((sxpar(hdr_ref, 'CD1_1'))^2. + (sxpar(hdr_ref, 'CD1_2'))^2.))
      endif
      pix_as_final = tab_pix_as_final(detarr)

      nstep_imprint = 10L    ;;; must be even and sufficiently large to ensure accurate imprint projection
      nstep_imprint_final = 10L
      diam_as_imprint = tab_diam_as_imprint(detarr)
      npass_crit = tab_npass_crit(detarr)

      file_signal = root_signal + arrays_default(detarr) + '.xdr'
      file_xyas = root_xyas + arrays_default(detarr) + '.xdr'
      file_xyas_final = root_xyas + 'final_' + arrays_default(detarr) + '.xdr'
      file_imprint = dir_grids + arrays_default(detarr) + '_imprintgrid_' + chain(pix_as, 2) + 'as_' $
         + chain(nstep_imprint) + '.xdr'
      file_imprint_final = dir_grids + arrays_default(detarr) + '_imprintgrid_' + chain(pix_as_final, 2) + 'as_' $
         + chain(nstep_imprint_final) + '.xdr'


      ;;; scale factor between coarse and fine grids:
      fact = ceil(fact_fwhm(detarr) * fwhm_as(detarr) / pix_as)
      pix_as_coarse = pix_as * fact
   endif



   restore, file_xyas_final

   ;;; map grid:
   if nblocks eq 1 then begin
      dimxas = float(max(xas, /nan) - min(xas, /nan)) + 2. * off_map_as
      dimyas = float(max(yas, /nan) - min(yas, /nan)) + 2. * off_map_as
      dimx = ceil(dimxas / pix_as)
      dimy = ceil(dimyas / pix_as)
      crpix1 = (float(off_cen_Xra * 3600.D0) + off_map_as) / pix_as
      crpix2 = (float(off_cen_Ydec * 3600.D0) + off_map_as) / pix_as
   endif

   if ((orient eq 'astro' or keyword_set(hdr_ref)) and (not keyword_set(spiral)) and nblocks eq 1) then begin
      maprot_deg = p_maprot_deg
      restore, file_xyas
      p_dimxas = float(max(xas, /nan) - min(xas, /nan)) + 2. * off_map_as
      p_dimyas = float(max(yas, /nan) - min(yas, /nan)) + 2. * off_map_as
      dimx = ceil(p_dimxas / pix_as)
      dimy = ceil(p_dimyas / pix_as)
   endif

   if !version.memory_bits ge 64 then begin
      dimx = long64(dimx)
      dimy = long64(dimy)
   endif


   ;;; pixel coordinates:
   x_bolos = float((off_map_as + xas) / pix_as)
   xas = 0
   y_bolos = float((off_map_as + yas) / pix_as)
   yas = 0
   save, file=file_pixcoords, x_bolos, y_bolos, /xdr
   x_bolos = 0
   y_bolos = 0



   if block eq block_start then begin

      ;;; number of time steps required to cross the FWHM:
      nt_cross = floor(pix_as_coarse / vscan_as_s / timestep)
      t_cross_fwhm = pix_as_coarse / vscan_as_s
      maxtimediff = 10. * nt_cross * timestep

      if nt_cross lt min_nsamples(detarr) then begin
         fact_fwhm = ceil(10. * min_nsamples * vscan_as_s * timestep / fwhm_as) / 10.
         w = where(fact_fwhm ge 1. and ind_arrays_eff gt 0, cw)
         if cw ge 1 then fact_scales(w) = 2. * fact_fwhm(w)
         print, crs + 'Value of the stability length:'
         for k = 0, narr - 1 do if ind_arrays_eff(k) gt 0 then $
            print, '     ' + chain(fact_fwhm(k), 1) + ' FWHM for ' + arrays_label(k)
         print, '   going back to recompute grids.' + crs
         vis_traject = 0
         goto, fact_fwhm_label
      endif


      ;;; dimensions for the coarse grid used in iterations:
      dimx_iter = dimx / fact
      if dimx mod fact gt 0 then dimx_iter += 1
      dimy_iter = dimy / fact
      if dimy mod fact gt 0 then dimy_iter += 1
   endif


   ;;; minimum number of samples per bolometer crossing:
   ntmin_pass = 999 + intarr(nt)
   restore, file_vscan
   w = where(vscan gt 0.)
   ntmin_pass(w) = floor(pix_as_coarse / vscan(w) / timestep)
   w = 0
   vscan = 0
   save, file=file_ntmin_pass, ntmin_pass, /xdr
   ntmin_pass = 0


   if block eq block_start then begin
      ;;; step of the coarse time grid:
      tc_min = fact_fwhm(detarr) * fwhm_as(detarr) / vscan_as_s
      tc_drift_s = max([tc_drift_s, tc_min])
      nt_aver = round(tc_drift_s / timestep)
      tc_drift_s = float(nt_aver * timestep)
      print, crs + 'Assuming for drifts a minimum timescale of ' + chain(tc_drift_s, 2) + ' s'
      print, 'Number of samples to cross a pixel of the coarse grid: ' + chain(nt_cross) + crs
   endif

   flag_grid_aver = 1
   tc_drift_s_ini = tc_drift_s

   flag_grid_aver = 2
   restore, file_times
   dt_sub = 220. * tc_drift_s
   for s = 0, nscans - 1 do begin
      wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1), csubs)
      if csubs ge 5 then dt_sub = min([dt_sub, median(time_s(ind_subscans(wsubs(1L:csubs-1L))-1L) - time_s(ind_subscans(wsubs(0L:csubs-2L))), /even)])
   endfor

   flag_nza = 0
   time_s = 0
   f_max = dt_sub / 2.2 / tc_drift_s
   n_sup = 3
   if f_max ge 81. then begin
      tab_nt_indiv = nt_aver * [81L, 27L, 9L, 3L, 1L + lonarr(n_sup)]
      n_lt_indiv = 4
   endif else if f_max ge 27. then begin
      tab_nt_indiv = nt_aver * [27L, 9L, 3L, 1L + lonarr(n_sup)]
      n_lt_indiv = 3
   endif else if f_max ge 9. then begin
      tab_nt_indiv = nt_aver * [9L, 3L, 1L + lonarr(n_sup)]
      n_lt_indiv = 2
   endif else if f_max ge 3. then begin
      tab_nt_indiv = nt_aver * [3L, 1L + lonarr(n_sup)]
      n_lt_indiv = 1
   endif else begin
      tab_nt_indiv = nt_aver * [1L + lonarr(n_sup)]
      n_lt_indiv = 0
   endelse
   iter_max = n_lt_indiv + n_sup

   if (nt / nt_aver ge max_nt_coarse and block eq block_start) then begin
      flag_grid_aver = 2
      nt_aver = ceil(float(nt) / float(max_nt_coarse))
      tc_drift_s_ini = float(nt_aver * timestep)
      print, crs + 'Increasing the minimum timescale for the average drift'
      print, 'in order to keep the number of coarse time steps below "max_nt_coarse".'
      print, '-> Assuming a new timescale of ' + chain(tc_drift_s_ini, 2) + ' s (for the average drift only)'
   endif


   ;;; coarse time grid made irregular in order not to mix edges of different scans:
   restore, file_times
   ;;; time_s, flag_nza
   flag_nza = 0

   ind_sc_coarse = [0L]
   w = where(time_s - time_s(0) le tc_drift_s_ini)
   ind_aver = [max(w) + 1]
   ws = 0
   repeat begin
      w = where(time_s(max(ind_aver):*) - time_s(max(ind_aver)) le tc_drift_s_ini)
      ind_aver_next = max(ind_aver) + max(w) + 1
      if ind_aver_next lt ind_scans(ws+1) then ind_aver = [ind_aver, ind_aver_next] $
         else begin
            ind_aver = [ind_aver, ind_scans(ws+1)]
            ind_sc_coarse = [ind_sc_coarse, n_elements(ind_aver)]
            ws += 1
         endelse
   endrep until (ws ge nscans)
   nt_save = n_elements(ind_aver)
   if !version.memory_bits ge 64 then nt_save = long64(nt_save)
   time_save = float((time_s([0, ind_aver(0:nt_save-2)]) + time_s(ind_aver-1)) / 2.)
   time_s = 0
   save, file=file_time_coarse, time_save, /xdr
   time_save = 0



   if block eq block_start then begin
      dim_imprint = 5L   ;;; dimension of the imprint image (must be odd)
      dim_imprint_final = max([dim_imprint, 2L + ceil(3. * fwhm_as(detarr) / 4. / pix_as_final)])
      if (dim_imprint_final mod 2L) eq 0 then dim_imprint_final += 1L
      xcmin_imprint = dim_imprint / 2
   endif


   save, file=file_scalars, file_times, file_time_coarse, file_signal, file_pixcoords, $
      file_projcoarse, file_projfine, file_imprint, file_maps, file_vscan, file_ntmin_pass, $
      inst, nt, nt_save, nt_aver, timestep, maxtimediff, nb, bref, nscans, nsubscans, $
      logf_inf, logf_sup, logf_high_inf, logf_high_sup, $
      vscan_as_s, dimx, dimy, dimx_iter, dimy_iter, fact, $
      pix_as, diam_as_imprint, dim_imprint, nstep_imprint, xcmin_imprint, $
      npass_crit, fact_npass_cen, visu, debug, /xdr


   if keyword_set(visu) then visu_arrays, file_scalars=file_scalars, ind_scans=ind_scans, elec_boxes=elec_boxes


   ;;; bolometer imprint matrix:
   if block eq block_start then compute_imprint, file_scalars=file_scalars


   ;;; estimate the noise level to be able to detect significant flux gradients:
   update_mask = 1
   uniformw = 0
   if keyword_set(visu) then window, 0, xs=1200
   compute_wnoise_artemis, file_scalars=file_scalars, maxnoise, highfnoise, update_mask=update_mask, $
      uniformw=uniformw, ind_scans=ind_scans, ind_subscans=ind_subscans, /initial, f_fwhm=fact_fwhm(detarr), $
      time_1=time_1, t_proc=t_proc
      print, crs + '"white noise" statistics (from spectral densities):'
      for i = 0, 9 do print, 'subscan ' + chain(i) + ':   min = ' $
         + chain(min(maxnoise(i, *), /nan), 5) + '   max = ' + chain(max(maxnoise(i, *), /nan), 5)
      print, '...'
      for i = nsubscans - 10, nsubscans - 1 do print, 'subscan ' + chain(i) + ':   min = ' $
         + chain(min(maxnoise(i, *), /nan), 5) + '   max = ' + chain(max(maxnoise(i, *), /nan), 5)


   print, crs + '"threshold noise":'
   for i = 0, 9 do print, 'subscan ' + chain(i) + ':   min = ' $
      + chain(min(highfnoise(i, *), /nan), 5) + '   max = ' + chain(max(highfnoise(i, *), /nan), 5)
   print, '...'
   for i = nsubscans - 10, nsubscans - 1 do print, 'subscan ' + chain(i) + ':   min = ' $
      + chain(min(highfnoise(i, *), /nan), 5) + '   max = ' + chain(max(highfnoise(i, *), /nan), 5)



   ;;; vectors for projection on fine grid:
   ;;; (to be done again every time missing samples are added)
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans


   ;;; initial maps (with and without flux cal. offsets):
   proj_map_nearest_artemis, map_simple, file_scalars=file_scalars, maxnoise=maxnoise, /todisk, $
      ind_subscans=ind_subscans

   map_init = proj_map_init_artemis(file_scalars=file_scalars, maxnoise=maxnoise, $
      ind_scans=ind_scans, ind_subscans=ind_subscans)



   ;;; vectors for projection on coarse grid:
   ;;; (to be done again every time missing samples are added)
   prepare_proj_coarse, file_scalars=file_scalars



   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      print, crs + 'initial signal spectral density (on scan length)' + crs
      compute_spd_signal_artemis, freq_comp, spd_comp_aver, file_scalars=file_scalars, $
         maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans
      time_1 = systime(/seconds)
   endif



   ;;;;;;;;;;;;;;;;;;;
   ;;; pre-processing
   ;;;;;;;;;;;;;;;;;;;


   ;;; initialize total drifts:
   drifts_all = fltarr(nt_save, nb)

;;;
;   file_mask = dir_in + 'mask_sources_' + file_out + '_' + arrays_default(detarr) + '.xdr'
;   decompose_artemis, file_scalars=file_scalars, ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, $
;      list_scanrot=list_scanrot, disjoint=disjoint, inner_edges=inner_edges, frac_overlap=frac_overlap, $
;      ind_overlap_scans=ind_overlap_scans, parity=parity, regions=regions, time_1=time_1, t_proc=t_proc, $
;      spiral=spiral
;;;




;;;***;;;
if (not keyword_set(nocorr)) then begin

   ;;; subtraction of baselines:

   ;;; to remove offsets due to inaccurate flux calibration,
   ;;; thermal drift to first order,
   ;;; and drifts with timescale similar to subscan duration

   print, crs + 'pre-processing: baseline subtraction (using simple fits)'
   baselines_artemis, drifts_all, map_back, size_deg, maprot_deg=maprot_deg, $
      file_scalars=file_scalars, ind_sc_coarse=ind_sc_coarse, $
      ind_aver=ind_aver, ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, $
      galactic=galactic, disjoint=disjoint, time_1=time_1, t_proc=t_proc



   proj_map_artemis, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans
   restore, file_maps
   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      disp_ima, map_signal, win=1, weightmap=map_weight, title='after baseline subtraction'
      disp_ima, map_simple, win=2, weightmap=map_weight, title='raw data'
      disp_ima, map_init - map_signal, win=3, weightmap=map_weight, $
         title='baselines (linear fits, without flux cal. offsets)'
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      time_1 = systime(/seconds)
   endif
   map_prev = map_signal
   map_signal = 0



;;;
if keyword_set(debug) then begin
   do_map_scans_nearest_artemis, map_scans, weight_scans, file_scalars=file_scalars, ind_scans=ind_scans, $
      ind_subscans=ind_subscans, maxnoise=maxnoise
   do_map_subarrays, map_subarr, weight_subarr, file_scalars=file_scalars, ind_scans=ind_scans, $
      ind_subscans=ind_subscans, maxnoise=maxnoise
   print, 'Control map_scans and map_subarr.'
   stop
endif
;;;

   map_prev = 0
   map_signal = 0
   map_weight = 0



   ;;; update white noise:
   if (not keyword_set(noglitch)) then mask_gl = 1 else mask_gl = 0
   if keyword_set(galactic) then map_back_noise = map_back else map_back_noise = 0
   compute_wnoise_artemis, file_scalars=file_scalars, maxnoise, highfnoise, ind_scans=ind_scans, $
      ind_subscans=ind_subscans, mask_gl=mask_gl, f_fwhm=fact_fwhm(detarr), map_back=map_back_noise, $
      time_1=time_1, t_proc=t_proc
   print, crs + 'updated "white noise" statistics:'
   for i = 0, 9 do print, 'subscan ' + chain(i) + ':   min = ' $
      + chain(min(maxnoise(i, *), /nan), 5) + '   max = ' + chain(max(maxnoise(i, *), /nan), 5)
   print, '...'
   for i = nsubscans - 10, nsubscans - 1 do print, 'subscan ' + chain(i) + ':   min = ' $
      + chain(min(maxnoise(i, *), /nan), 5) + '   max = ' + chain(max(maxnoise(i, *), /nan), 5)
   print, crs


   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      print, crs + 'signal spectral density after baseline subtraction (on scan length)' + crs
      compute_spd_signal_artemis, freq_comp, spd_comp_aver, file_scalars=file_scalars, $
         maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans
      time_1 = systime(/seconds)
   endif



   ;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;; drift subtraction loop
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;***;;;
if (not keyword_set(nodrifts)) then begin

   iter = 0
   iter_indiv = 0
   indiv_completed = 0
   amp_drifts = 999. * maxnoise

   aver_completed = 0


   repeat begin

      restore, file_scalars

      if (aver_completed eq 1 and iter_indiv eq 0) then begin
         drifts_all_aver = drifts_all
         restore, file_time_coarse
         time_save_aver = time_save
         time_save = 0
         drifts_all = fltarr(nt_save, nb)
      endif

      ;;; update coarse time grid after average drift subtraction:
      if (aver_completed eq 1 and flag_grid_aver eq 2) then begin
         restore, file_scalars
         nt_aver_old = nt_aver
         restore, file_time_coarse
         time_save_old = time_save
         drifts_all_old = drifts_all
         nt_aver = tab_nt_indiv(iter_indiv)
         tc_drift_s = float(nt_aver * timestep)

         restore, file_times
         flag_nza = 0
         ind_sc_coarse = [0L]
         w = where(time_s - time_s(0) le tc_drift_s)
         ind_aver = [max(w) + 1]
         ws = 0
         repeat begin
            w = where(time_s(max(ind_aver):*) - time_s(max(ind_aver)) le tc_drift_s)
            ind_aver_next = max(ind_aver) + max(w) + 1
            if ind_aver_next lt ind_scans(ws+1) then ind_aver = [ind_aver, ind_aver_next] $
               else begin
               ind_aver = [ind_aver, ind_scans(ws+1)]
               ind_sc_coarse = [ind_sc_coarse, n_elements(ind_aver)]
               ws += 1
            endelse
         endrep until (ws ge nscans)
         nt_save = n_elements(ind_aver)
         if !version.memory_bits ge 64 then nt_save = long64(nt_save)
         time_save = float((time_s([0, ind_aver(0:nt_save-2)]) + time_s(ind_aver-1)) / 2.)
         time_s = 0
         save, file=file_time_coarse, time_save, /xdr

         drifts_all = fltarr(nt_save, nb)
         for b = 0L, nb - 1 do drifts_all(*, b) = interpol(drifts_all_old(*, b), time_save_old, time_save)
         time_save_old = 0
         drifts_all_old = 0
         time_save = 0

         if iter_indiv eq n_lt_indiv then flag_grid_aver = 1
      endif
      save, file=file_scalars, file_times, file_time_coarse, file_signal, file_pixcoords, $
         file_projcoarse, file_projfine, file_imprint, file_maps, file_vscan, file_ntmin_pass, $
         inst, nt, nt_save, nt_aver, timestep, maxtimediff, nb, bref, nscans, nsubscans, $
         logf_inf, logf_sup, logf_high_inf, logf_high_sup, $
         vscan_as_s, dimx, dimy, dimx_iter, dimy_iter, fact, $
         pix_as, diam_as_imprint, dim_imprint, nstep_imprint, xcmin_imprint, $
         npass_crit, fact_npass_cen, visu, debug, /xdr


      ;;; update white noise and/or mask glitches:
      mask_gl = 0
      if (iter eq 1 or iter eq 3) then begin
         if keyword_set(galactic) then map_back_noise = map_back else map_back_noise = 0
         compute_wnoise_artemis, file_scalars=file_scalars, maxnoise, highfnoise, ind_scans=ind_scans, $
            ind_subscans=ind_subscans, mask_gl=mask_gl, f_fwhm=fact_fwhm(detarr), map_back=map_back_noise, $
            time_1=time_1, t_proc=t_proc
         print, crs + 'updated "white noise" statistics:'
         for i = 0, 9 do print, 'subscan ' + chain(i) + ':   min = ' $
            + chain(min(maxnoise(i, *), /nan), 5) + '   max = ' + chain(max(maxnoise(i, *), /nan), 5)
         print, '...'
         for i = nsubscans - 10, nsubscans - 1 do print, 'subscan ' + chain(i) + ':   min = ' $
            + chain(min(maxnoise(i, *), /nan), 5) + '   max = ' + chain(max(maxnoise(i, *), /nan), 5)
         print, crs
      endif


      if (keyword_set(debug) and iter gt 0) then begin
         time_2 = systime(/seconds)
         t_proc += (time_2 - time_1)
         print, crs + 'signal spectral density after iteration ' + chain(iter-1) + crs
         compute_spd_signal_artemis, freq_comp, spd_comp_aver, file_scalars=file_scalars, $
            maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans
         time_1 = systime(/seconds)
      endif


      if aver_completed eq 0 then mode_drifts = 'average drift' else mode_drifts = 'individual drifts'
      min_nsamp = min_nsamples(detarr)



      ;;; new projection:
      if (aver_completed eq 0 or iter_indiv gt n_lt_indiv - 2) then proj_map_artemis, $
         file_scalars=file_scalars, maxnoise=maxnoise, ind_subscans=ind_subscans
      if keyword_set(visu) then begin
         time_2 = systime(/seconds)
         t_proc += (time_2 - time_1)
         restore, file_maps
         disp_ima, map_signal, win=1, title='iteration '+chain(iter)+' : input map', weightmap=map_weight
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
         map_signal = 0
         map_weight = 0
         time_1 = systime(/seconds)
      endif



      ;;; drift subtraction:

      if mode_drifts eq 'average drift' then print, 'iteration ' + chain(iter) + ' : ' + mode_drifts $
       else begin
;;;         if (iter_indiv le n_lt_indiv and (iter_indiv mod 2) eq 0) then do_elec_bands = 1 else do_elec_bands = 0
;;;         if iter_indiv ge n_lt_indiv then do_elec_bands = 0
         do_elec_bands = 0
;;;
         if do_elec_bands eq 1 then label_drifts = 'per-electronic-box drifts (single config. for all scans)' $
            else label_drifts = mode_drifts
         print, 'iteration ' + chain(iter) + ' : ' + label_drifts
       endelse

      if aver_completed lt 1 then print, '              time resolution = ' + chain(tc_drift_s_ini, 2) + ' s' $
         else print, '              time resolution = ' + chain(tc_drift_s, 2) + ' s'

      detect_gl = 0


      if mode_drifts eq 'average drift' then begin

         title = 'iteration ' + chain(iter) + ' : signal & drifts (all-bolometers average)'
         dr_bolo = p_arrhw(detarr) * pfov(detarr) / pix_as
         do_elec_bands = 0

         amp_drifts = subtract_aver_drift_artemis(drifts_all, flag_valid_scans=flag_valid_scans, $
            maxnoise=maxnoise, highfnoise=highfnoise, ind_sc_coarse=ind_sc_coarse, ind_scans=ind_scans, $
            ind_subscans=ind_subscans, ind_aver=ind_aver, file_scalars=file_scalars, title=title, $
            min_nsamp=min_nsamp, detect_gl=detect_gl, t_cross_fwhm=t_cross_fwhm, f_fwhm=fact_fwhm(detarr), $
            iter=iter, galactic=galactic, size_deg=size_deg, dr_bolo=dr_bolo, time_1=time_1, t_proc=t_proc)

      endif else begin

         title = 'iteration ' + chain(iter) + ' : signal & drifts (bolometer ' + chain(bref) + ')'
         after_indiv = 1

         amp_drifts = subtract_indiv_drifts_artemis(drifts_all, stopcorr, maxnoise=maxnoise, highfnoise=highfnoise, $
            ind_sc_coarse=ind_sc_coarse, ind_scans=ind_scans, ind_subscans=ind_subscans, ind_aver=ind_aver, $
            file_scalars=file_scalars, title=title, min_nsamp=min_nsamp, detect_gl=detect_gl, $
            t_cross_fwhm=t_cross_fwhm, f_fwhm=fact_fwhm(detarr), iter_tot=iter, iter_indiv=iter_indiv, $
            time_1=time_1, t_proc=t_proc, n_lt_indiv=n_lt_indiv, $
            do_elec_bands=do_elec_bands, elec_boxes=elec_boxes)
      endelse




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; redo baseline subtraction after average drift subtraction

if (mode_drifts eq 'average drift' or iter_indiv le max([0, n_lt_indiv - 3])) then begin

   proj_map_artemis, file_scalars=file_scalars, maxnoise=maxnoise, ind_subscans=ind_subscans

   print, crs + 'new iteration of baseline subtraction'
   restore, file_maps
   map_prev = map_signal


   force_nsig=2.
   if (keyword_set(galactic) and iter_indiv eq max([0, n_lt_indiv - 3])) then cube_mask = 1 else cube_mask = 0
   baselines_artemis, drifts_all, map_back, size_deg, maprot_deg=maprot_deg, $
      file_scalars=file_scalars, ind_sc_coarse=ind_sc_coarse, $
      ind_aver=ind_aver, ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, $
      galactic=galactic, time_1=time_1, t_proc=t_proc, force_nsig=force_nsig, after_indiv=after_indiv, $
      cube_mask=cube_mask, disjoint=disjoint


   proj_map_artemis, file_scalars=file_scalars, maxnoise=maxnoise, ind_subscans=ind_subscans

   restore, file_maps
   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      disp_ima, map_signal, win=1, weightmap=map_weight, title='after drift + baseline subtraction'
      disp_ima, map_prev - map_signal, win=3, weightmap=map_weight, $
         title='baselines (linear fits, without flux cal. offsets)'
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      time_1 = systime(/seconds)
   endif
   map_prev = map_signal
endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



      ;;; prepare next iteration:
      iter += 1
      if mode_drifts eq 'individual drifts' then iter_indiv += 1

      wdef = where(finite(maxnoise) gt 0, cdef)
      wconv = where(amp_drifts(wdef) / maxnoise(wdef) lt 1., cconv)
      if ((float(cconv) / float(cdef) ge 0.97 and (median(amp_drifts / maxnoise, /even) le 0.7 $
         or max(amp_drifts / maxnoise, /nan) lt 1.)) or (median(amp_drifts / maxnoise, /even) le 0.5 $
         and max(amp_drifts / maxnoise, /nan) lt 1.3)) then begin
            if (mode_drifts eq 'individual drifts' and (iter gt 1 or max(amp_drifts) eq 0.) $
               and iter_indiv gt n_lt_indiv) then indiv_completed = 1
            if mode_drifts eq 'average drift' then aver_completed = 1
      endif
      if (aver_completed lt 1 and iter ge iter_aver_max) then aver_completed = 1
      if (keyword_set(stopcorr) and mode_drifts eq 'individual drifts') then indiv_completed = 1

   endrep until (iter_indiv ge iter_max or (indiv_completed eq 1 and aver_completed eq 1))



   
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   print, crs + 'data processed in ' + chain(t_proc) + ' s'
   print, 'observation duration = ' + chain(nt * timestep) + ' s'



   restore, file_time_coarse
   for b = 0L, nb - 1L do drifts_all(*, b) += interpol(drifts_all_aver(*, b), time_save_aver, time_save)
   drifts_all_aver = 0
   time_save_aver = 0
   time_save = 0



   if keyword_set(visu) then begin
      print, crs + 'signal spectral density after full correction'
      print, crs + ' (on scan length if continuous data stream ; otherwise on subscan length)'
      compute_spd_signal_artemis, freq_comp, spd_comp_aver, file_scalars=file_scalars, $
         maxnoise=maxnoise, ind_scans=ind_scans, ind_subscans=ind_subscans
      freq_comp = 0
      spd_comp = 0
      spd_comp_aver = 0

      ;;; compute spectral density of removed low-frequency noise:
      print, 'subtracted drifts spectral density (on coarse time grid)'
      compute_nspd_artemis, file_scalars=file_scalars, drifts_sum=drifts_all, maxnoise=maxnoise, $
         ind_scans=ind_scans, ind_subscans=ind_subscans, $
         title='total drifts spectral density', ind_sc_coarse=ind_sc_coarse
   endif



   ;;; respective amplitudes of thermal drift and individual drifts:

   drift_thermal = fltarr(nt_save)
   drift_th_nobase = fltarr(nt_save)
   restore, file_time_coarse
   restore, file_times
   for s = 0L, nscans - 1L do begin
      wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1) - 1L, csubs)
      for subs = wsubs(0L), wsubs(csubs-1L) do begin
         wb = where(finite(maxnoise(subs, *)) gt 0, cb)
         wt = where(ind_aver ge ind_subscans(subs) and ind_aver lt ind_subscans(subs+1), ct)
         if (s eq nscans - 1L and subs eq wsubs(csubs-1L)) then wt = where(ind_aver ge ind_subscans(subs), ct)
         drifts_scan = drifts_all(wt, *)
         drift_thermal(wt) = total(drifts_scan(*, wb) * rebin(finite(maxnoise(subs, wb)), ct, cb), 2) $
            / total(finite(maxnoise(subs, wb)))
      endfor
      drifts_scan = 0
      wt = where(ind_aver ge ind_scans(s) and ind_aver lt ind_scans(s+1), ct)
      if s eq nscans - 1L then wt = where(ind_aver ge ind_scans(s), ct)
      off = min(drift_thermal(wt))
      drift_thermal(wt) -= off
      drifts_all(wt, *) -= off
      drift_th_nobase(wt) = drift_thermal(wt) - absdev_fit(time_save(wt), drift_thermal(wt), $
         noise_level=30.*median(maxnoise), /sym, /dyn, /nomask)
      drift_th_nobase(wt) -= min(drift_th_nobase(wt))
   endfor

   if keyword_set(debug) then begin
      wset, 0
      otherplot, time_save, drift_thermal, xtitle='time (s)', ytitle='thermal drift'
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   endif

   med = mean(drift_thermal)
   amp_thermal = 3. * sqrt(mean((drift_thermal - med)^2.))

   drifts_uncorr = drifts_all - rebin(drift_thermal, nt_save, nb)
   w = where(abs(drifts_all) lt (Machar()).eps or finite(drifts_all) lt 1, cw)
   if cw ge 1 then drifts_uncorr(w) = 0.
   for i = 0, nscans - 1 do begin
      wt = where(ind_aver ge ind_scans(i) and ind_aver lt ind_scans(i+1), ct)
      if i eq nscans - 1L then wt = where(ind_aver ge ind_scans(i), ct)
      wsubs = where(ind_subscans ge ind_scans(i) and ind_subscans lt ind_scans(i+1) - 1L)
      for b = 0L, nb - 1L do if total(finite(maxnoise(wsubs, b)), 1) gt 0 then begin
         tab = drifts_uncorr(wt, b)
         drifts_uncorr(wt, b) -= mean(tab(where(abs(tab) gt (Machar()).eps)))
      endif
   endfor
   wb = where(total(finite(maxnoise), 1) ge max(total(finite(maxnoise), 1)) / 4L)
   tab = drifts_uncorr(*, wb)
   amp_uncorr = 3. * sqrt(mean((tab(where(abs(tab) gt (Machar()).eps)))^2.))
   tab = 0

   time_s = 0
   flag_nza = 0
   time_save = 0

   print, crs + 'drift amplitudes in units of median white noise and in Jy/bolometer:'
   print, '     thermal: ' + chain(amp_thermal / median(maxnoise), 2) + ' WN or ' $
      + chain(amp_thermal, 5) + ' Jy/b'
   print, '     uncorrelated: ' + chain(amp_uncorr / median(maxnoise), 2) + ' WN or ' $
      + chain(amp_uncorr, 5) + ' Jy/b'


   if keyword_set(visu) then begin
      compute_nspd_artemis, file_scalars=file_scalars, $
         drifts_sum=rebin(drift_thermal, nt_save, nb), maxnoise=maxnoise, ind_scans=ind_scans, $
         ind_subscans=ind_subscans, title='average drift spectral density', $
         ind_sc_coarse=ind_sc_coarse
      compute_nspd_artemis, file_scalars=file_scalars, drifts_sum=drifts_uncorr, maxnoise=maxnoise, $
         ind_scans=ind_scans, ind_subscans=ind_subscans, $
         title='uncorrelated drifts spectral density', ind_sc_coarse=ind_sc_coarse
   endif

;;;***;;;
endif
endif




   ;;;;;;;;;;;;;;;;;;;;;
   ;;; final projection
   ;;;;;;;;;;;;;;;;;;;;;

   if (pix_as_final ne pix_as or orient eq 'astro' or keyword_set(hdr_ref)) then begin

      restore, file_xyas_final

      if keyword_set(hdr_ref) then begin
         x_bolos = float(xas / pix_as_final)
         xas = 0
         y_bolos = float(yas / pix_as_final)
         yas = 0
      endif else begin
         x_bolos = float((off_map_as + xas) / pix_as_final)
         xas = 0
         y_bolos = float((off_map_as + yas) / pix_as_final)
         yas = 0
      endelse
      save, file=file_pixcoords, x_bolos, y_bolos, /xdr
      x_bolos = 0
      y_bolos = 0

      diameter = 2. * sqrt(0.5) * pfov(detarr) / sqrt(!pi)
      if block gt block_start then update_only = 1 else update_only = 0
      compute_imprint, file_scalars=file_scalars, pixsize=pix_as_final, $
         diameter=diameter, dim_imprint_final=dim_imprint_final, nstep_imprint_final=nstep_imprint_final, $
         file_imprint_final=file_imprint_final, update_only=update_only

      if nzadata ne 'no' then prepare_proj_fine, file_scalars=file_scalars, $
         xdim=dimx_final, ydim=dimy_final, ind_scans=ind_scans

   endif else begin
      restore, file_scalars
      dimx_final = dimx
      dimy_final = dimy
      crpix1_final = crpix1
      crpix2_final = crpix2
   endelse


   ;;; include or not non-zero acceleration data:
   if nzadata eq 'no' then prepare_proj_fine, file_scalars=file_scalars, $
      xdim=dimx_final, ydim=dimy_final, ind_scans=ind_scans, nzadata=nzadata


   print, crs + 'final projection of drift-corrected time series'

   restore, file_times
   flag_nza = 0
   restore, file_time_coarse
   drifts = fltarr(nt, nb, 3)
   for b = 0L, nb - 1 do drifts(*, b, 0) = interpol(drifts_all(*, b), time_save, float(time_s))
;;;   if keyword_set(drift_th_nobase) then drifts(*, *, 1) = rebin(interpol(drift_th_nobase, time_save, float(time_s)), nt, nb)
   if keyword_set(drift_thermal) then drifts(*, *, 1) = rebin(interpol(drift_thermal, time_save, float(time_s)), nt, nb)
;;;
   if keyword_set(drifts_uncorr) then for b = 0L, nb - 1 do drifts(*, b, 2) = interpol(drifts_uncorr(*, b), time_save, float(time_s))
   time_s = 0
   time_save = 0

   proj_map_artemis, file_scalars=file_scalars, nstep_imprint_final=nstep_imprint_final, $
      file_imprint_final=file_imprint_final, dimx_proj=dimx_final, dimy_proj=dimy_final, $
      maxnoise=maxnoise, ind_scans=ind_scans, /comp_errors, lowfnoise=drifts, ind_subscans=ind_subscans, $
      cube_mask=cube_mask
   drifts = 0
   cube_mask = 0

   restore, file_maps


   ;;; set approximately the zero level of the map:
   stat_map_artemis, med, sig, map=map_signal, weightmap=map_weight, noise_level=median(maxnoise)
   if sig gt 0.3 * median(maxnoise) then begin
       w = where(map_weight gt 0.)
       med_weight = median(map_weight(w))
       w = where(map_weight ge 0.4 * med_weight, cw)
       histo = histogram(map_signal(w), locations=sb, binsize=median(maxnoise)/5., /nan)
       w = where(total(histo, /cumulative) ge cw / 50.)
       med = sb(w(0))
   endif
   w = where(map_weight gt 0.)
   map_signal(w) -= med

   if keyword_set(visu) then begin
      disp_ima, map_signal, win=1, title='final map', weightmap=map_weight
      disp_ima, map_weight, win=2, title='weight map'
      disp_ima, map_error, win=3, title='error map', weightmap=map_weight, noise_level=median(maxnoise)/10.
      disp_ima, map_lowfnoise, win=4, title='total drifts (excluding flux cal. offsets)', $
        weightmap=map_weight, noise_level=10.*median(maxnoise)
      disp_ima, map_signal / map_error, win=5, min_map=2., title='naive SNR map', weightmap=map_weight
;;;      disp_ima, map_lfn_2comp(*, *, 0), win=6, title='thermal drift (excluding scan baselines)', weightmap=map_weight
      disp_ima, map_lfn_2comp(*, *, 0), win=6, title='thermal drift (excluding scan offsets)', weightmap=map_weight
;;;
      disp_ima, map_lfn_2comp(*, *, 1), win=7, title='uncorrelated drifts (excluding scan offsets)', weightmap=map_weight
      if keyword_set(map_mask) then disp_ima, map_mask, win=8, title='source mask'
      res = dialog_message('Write output ?', /center, /information, title='Pause')
   endif




;;;;;;;;;;;;;;;;;;;
;;; output written
;;;;;;;;;;;;;;;;;;;


   ;;; reinject data into original structures:


   if keyword_set(dir_out_processed) then begin
      print, crs+'Preparing to re-inject processed data into subscan structures:'
      print, '   donnees_red will be overwritten.'
      print, crs+'The "otf_subscan_#.xdr" files are expected to be found in:'
      print, dir_out_processed
      res = 'Yes'
      if keyword_set(visu) then res = dialog_message('Overwrite donnees_red ?', /center, /question)
      if res eq 'Yes' then begin
         list_scannum = long(stregex(files_scans, '[0-9]{5,10}', /extract))
         restore, file_signal
         for s = 0, nscans - 1 do convert_input_subscans_artemis, $
            dir_out=dir_out_default, scannum=list_scannum(s), array=band, $
            /reverse, signal_scan=signal(ind_scans(s):ind_scans(s+1)-1, *), debug=debug
         signal = 0
      endif
   endif


   nplanes = 6
   if keyword_set(map_mask) then nplanes += 1
   cube = fltarr(dimx_final, dimy_final, nplanes)
   cube(*, * ,0) = map_signal
   cube(*, * ,1) = map_error
   cube(*, *, 2) = map_lowfnoise
   cube(*, *, 3) = map_weight
   cube(*, *, 4:5) = map_lfn_2comp
   if keyword_set(map_mask) then cube(*, *, 6) = map_mask
   w = where(map_weight le 0., cw)
   if cw ge 1 then for i = 0, nplanes - 1 do if i ne 3 then cube(w+dimx_final*dimy_final*i) = !values.f_nan

   ;;; brightness unit:
   bunit = 'Jy/beam'

   key_planes = 'PLANE' + chain(indgen(nplanes))
   str_planes = ['signal map', 'error map', 'total drifts (excluding flux calibration offsets)', 'weight map']
;;;   str_planes = [str_planes, 'average drift (excluding scan baselines)', 'uncorrelated drifts (excluding scan offsets)']
   str_planes = [str_planes, 'average drift (excluding scan offsets)', 'uncorrelated drifts (excluding scan offsets)']
;;;
   if keyword_set(map_mask) then str_planes = [str_planes, 'source mask (fraction of samples masked)']

   mkhdr, hdr, cube
   sxaddpar, hdr, 'TELESCOP', 'APEX'
   sxaddpar, hdr, 'INSTRUME', strupcase(inst)
   sxaddpar, hdr, 'FILTER', arrays_names(detarr)
   sxaddpar, hdr, 'OBSMODE', 'SCAN'
   sxaddpar, hdr, 'VSCAN1', round(100. * vs_min) / 100., 'lower scan speed (arcsec/s)'
   sxaddpar, hdr, 'VSCAN2', round(100. * vs_max) / 100., 'upper scan speed (arcsec/s)'
   sxaddpar, hdr, 'NSCANS', nscans, 'number of combined scans'
   sxaddpar, hdr, 'NBLOCKS', nblocks, 'number of spatial blocks for processing'
   sxaddpar, hdr, 'BLOCK', 'TOTAL', 'block number of this map / total map'
   sxaddpar, hdr, 'BUNIT', bunit
   sxaddpar, hdr, 'PFOV', pix_as_final, 'pixel size (arcseconds)'
   sxaddpar, hdr, 'CTYPE1', 'RA---TAN'
   sxaddpar, hdr, 'CTYPE2', 'DEC--TAN'
   sxaddpar, hdr, 'EQUINOX', 2000.
   sxaddpar, hdr, 'CRPIX1', crpix1_final + 0.5, 'reference pixel in X coordinate'
   sxaddpar, hdr, 'CRPIX2', crpix2_final + 0.5, 'reference pixel in Y coordinate'
   sxaddpar, hdr, 'CRVAL1', crval1, 'RA of reference position (degrees)'
   sxaddpar, hdr, 'CRVAL2', crval2, 'DEC of reference position (degrees)'
   cdelt = pix_as_final / 3600.
   sxaddpar, hdr, 'CDELT1', -cdelt
   sxaddpar, hdr, 'CDELT2', cdelt
   if orient eq 'astro' then crota2 = 0.D0 else crota2 = -scanrot_deg
   sxaddpar, hdr, 'CROTA2', crota2, 'angle from Y axis to North axis (degrees)'
   for i = 0, nplanes - 1 do sxaddpar, hdr, key_planes(i), str_planes(i)
   sxaddpar, hdr, 'SOFTWARE', 'L1-L2: Scanamorphos for Artemis '+vers+' (cf 2013PASP..125.1126R)'
   if keyword_set(galactic) then options = 'galactic' else options = ''
   if keyword_set(noglitch) then if options eq '' then options = 'noglitch' $
      else options = options + ', noglitch'
   if options ne '' then sxaddpar, hdr, 'OPTIONS', options else sxaddpar, hdr, 'OPTIONS', 'none'


   hdr_block = hdr
   if nblocks gt 1 then sxaddpar, hdr_block, 'BLOCK', block, 'block number of this map / total map'

   if (nblocks eq 1 and keyword_set(one_plane_fits)) then begin
      hdr_ima = hdr_block
      sxaddpar, hdr_ima, 'NAXIS', 2
      sxdelpar, hdr_ima, 'NAXIS3'
      sxdelpar, hdr_ima, key_planes
      suffix = ['signal', 'error', 'drifts_tot', 'weight', 'averdrift', 'uncorrdrifts']
      if keyword_set(map_mask) then suffix = [suffix, 'source_mask']
      for i = 0, nplanes - 1 do begin
         hdr_plane = hdr_ima
         sxaddpar, hdr_plane, key_planes(i), str_planes(i)
         writefits, dir_out + file_out + '_' + inst + arrays_names(detarr) + '_' + suffix(i) + '.fits', $
            cube(*, *, i), hdr_plane
      endfor
   endif else writefits, dir_out + file_out + '_' + inst + arrays_names(detarr) + '_' + chain(block) + '.fits', $
      cube, hdr_block


   if keyword_set(visu) then begin
      print, crs + 'time for last checks'
      print, 'Now is the time to save the corrected time series, and/or other variables'
      print, '(move the files "save_signal_[detector].xdr" to a subdirectory).'
      res = dialog_message('   Finish ?   ', /center, /information, title='Pause')
   endif

   while !d.window ge 0 do wdelete, !d.window

   cube = 0
   map_signal = 0
   map_error = 0
   map_lowfnoise = 0
   map_weight = 0
   map_transient = 0
   map_lfn_2comp = 0
   map_mask = 0

endif

endfor


;;; clean up intermediate files written to disk:
if (not keyword_set(save_final)) then begin
   file_delete, file_scalars, file_times, file_time_coarse, file_projcoarse, file_projfine, $
      file_pixcoords, file_vscan, file_ntmin_pass, file_maps, $
      file_XYscan, file_param_blocks, /noexpand_path, /quiet
   for detarr = 0, narr - 1 do if ind_arrays_eff(detarr) eq 1 then begin
      suffix = arrays_default(detarr) + '.xdr'
      file_delete, root_signal + suffix, root_xyas + suffix, root_xyas + 'final_' + suffix, $
         dir_grids + arrays_default(detarr) + '_imprintgrid_*.xdr', /noexpand_path, /quiet
   endif
endif




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; blocks stitched together 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

if nblocks gt 1 then begin
   print, crs + 'Combining the ' + chain(nblocks) + ' spatial blocks into a single map.' + crs

   w = where(ind_arrays_eff eq 1)
   ind = w(0)
   root_out = dir_out + file_out + '_' + inst + arrays_names(ind)
   stitch_blocks, dimx_final=dimx_final, dimy_final=dimy_final, nplanes=nplanes, nblocks=nblocks, $
      plane_weight=3, root_out=root_out, hdr=hdr, indep='no'
endif


while !d.window ge 0 do wdelete, !d.window

print, crs + 'Bye' + crs


end_label:
end


;;;;;;;;;
;;;;;;;;;
;;;;;;;;;
