function subtract_indiv_drifts_artemis, drifts_all, stopcorr, maxnoise=maxnoise, highfnoise=highfnoise, $
         ind_sc_coarse=ind_sc_coarse, ind_scans=ind_scans, ind_subscans=ind_subscans, ind_aver=ind_aver, $
         file_scalars=file_scalars, title=title, min_nsamp=min_nsamp, detect_gl=detect_gl, $
         t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, iter_tot=iter_tot, iter_indiv=iter_indiv, $
         time_1=time_1, t_proc=t_proc, n_lt_indiv=n_lt_indiv, $
         do_elec_bands=do_elec_bands, elec_boxes=elec_boxes


;;; computes individual drifts, interpolates and subtracts them from the signal

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

crs = string(13b) + string(10b)


compute_indiv_drifts_artemis, array_drifts, array_npass, w_glitches, nb_glitch, $
   maxnoise=maxnoise, highfnoise=highfnoise, ind_scans=ind_scans, ind_aver=ind_aver, $
   file_scalars=file_scalars, min_nsamp=min_nsamp, detect_gl=detect_gl, $
   t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, time_1=time_1, t_proc=t_proc, $
   iter_indiv=iter_indiv, n_lt_indiv=n_lt_indiv, ind_subscans=ind_subscans



if do_elec_bands eq 1 then begin
   for s = 0, nscans - 1 do for bl = min(elec_boxes(s, *), /nan), max(elec_boxes(s, *), /nan) do begin
      wb = where(elec_boxes(s, *) eq bl, cb)
      if cb ge 1 then begin
         w = where(total(array_npass(ind_sc_coarse(s):ind_sc_coarse(s+1)-1L, wb), 2) gt cb / 3L, cw, complement=wz, ncomplement=cz)
         if cw ge 1 then begin
            dr = array_drifts(ind_sc_coarse(s)+w, *)
            np = array_npass(ind_sc_coarse(s)+w, *)
            dr(*, wb) = rebin(total(dr(*, wb) * np(*, wb), 2, /nan) / total(np(*, wb), 2, /nan), cw, cb)
            np(*, wb) = rebin(total(np(*, wb), 2), cw, cb)
            array_drifts(ind_sc_coarse(s)+w, *) = dr
            array_npass(ind_sc_coarse(s)+w, *) = np
         endif
      endif
   endfor
endif




;;; mask glitches and update projection vectors:
if nb_glitch ge 1 then begin
   print, crs + 'removed ' + chain(nb_glitch) + ' sample(s) affected by glitches'
   restore, file_signal
   signal(w_glitches) = !values.f_nan
   save, file=file_signal, signal, /xdr
   signal = 0

   prepare_proj_coarse, file_scalars=file_scalars
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
   w_glitches = 0

   if (keyword_set(visu) and nb_glitch gt 100) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      restore, file_maps
      proj_map_nearest_artemis, map_noglitch, file_scalars=file_scalars, maxnoise=maxnoise, ind_subscans=ind_subscans
      disp_ima, map_signal - map_noglitch, win=4, title='map of residual glitches', $
         weightmap=map_weight, /trunc_max, noise_level=median(maxnoise)
      res = dialog_message('Continue and delete window ?', /center, /information, title='Pause')
      wdelete, 4
      time_1 = systime(/seconds)
   endif
endif



restore, file_times

restore, file_time_coarse


drifts = fltarr(nt, nb)
amp_drifts = fltarr(nsubscans, nb)



npass_min = 0.


;;; stop correction after 3 iterations if it is sparsely determined:

flag_nza_coarse = interpol(flag_nza, float(time_s), time_save)
stopcorr = 0
if (not keyword_set(n_lt_indiv)) then n_lt_indiv = 0
if iter_indiv ge max([2, n_lt_indiv + 2]) then begin
   wnom = where(flag_nza_coarse gt 0., cnom)
   wdef = where(rebin(flag_nza_coarse, nt_save, nb) gt 0. and array_npass gt npass_min, cdef)
   wb = where(finite(maxnoise) gt 0, cb)
   cdef /= (cb / float(nsubscans))
   if cdef lt 0.5 * cnom then begin
      print, '              drifts sparsely determined -> correction to be stopped after current iteration.'
      stopcorr = 1
   endif
endif


wdef = where(array_npass gt npass_min)
med = median(array_drifts(wdef))

npass_min = fltarr(nscans)



for i = 0, nscans - 1 do begin
   wsubs = where(ind_subscans ge ind_scans(i) and ind_subscans le ind_scans(i+1) - 1, csubs)
   for b = 0L, nb - 1L do if total(finite(maxnoise(wsubs, b)), 1) gt 0. then begin
      time_ref = time_s(ind_scans(i))
      csave = ind_sc_coarse(i+1) - ind_sc_coarse(i)
      wsave = ind_sc_coarse(i) + lindgen(csave)
;;;
;      wz = where((array_npass(wsave, b) le npass_min(i) and flag_nza_coarse(wsave) gt 0. $
;         and array_npass(wsave, b) lt 0.65 * smooth(array_npass(wsave, b), 7)) or array_npass(wsave, b) le 0. $
      wz = where((array_npass(wsave, b) le npass_min(i) and flag_nza_coarse(wsave) gt 0.) or array_npass(wsave, b) le 0. $
;;;
         or finite(array_drifts(wsave, b)) lt 1, cz, complement=wdef, ncomplement=cdef)
      if (cdef ge 3 and (iter_indiv lt n_lt_indiv or cdef ge 15)) then begin
         if cz ge 1 then array_drifts(wsave(wz), b) = 0.
         tab_time = [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(i+1)-1) + timestep - time_ref)]
         tab_drifts = [array_drifts(wsave(0), b), array_drifts(wsave, b), array_drifts(wsave(csave-1), b)]
         subs_drifts = interpol(tab_drifts, tab_time, float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))
         drifts(ind_scans(i):ind_scans(i+1)-1, b) = subs_drifts
         for isubs = 0L, csubs - 1L do begin
            subs = wsubs(isubs)
            w = where(flag_nza(ind_subscans(subs):ind_subscans(subs+1)-1) gt 0, cw)
            cut_drifts = subs_drifts(ind_subscans(subs)-ind_scans(i):ind_subscans(subs+1)-1-ind_scans(i))
            if cw lt (ind_subscans(subs+1) - ind_subscans(subs)) / 3. then dev = sqrt(mean((cut_drifts - med)^2.)) $
               else dev = sqrt(mean((cut_drifts(w) - med)^2.))
            amp_drifts(subs, b) = 3. * dev
         endfor
      endif else begin
         array_drifts(wsave, b) = 0.
         drifts(ind_scans(i):ind_scans(i+1)-1, b) = 0.
      endelse
   endif
endfor


;;; update total drifts:
drifts_all += array_drifts


restore, file_signal


;;; visualize result for reference bolometer:
if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   window, 0, xs=1200
   wdev = where(finite(signal(*, bref)) gt 0., cw)
   while cw lt 10 do begin
      bref += 1
      wdev = where(finite(signal(*, bref)) gt 0., cw)
   endwhile
   med = median(signal(wdev, bref))
   dev = sqrt(mean((signal(wdev, bref) - med)^2.))
   for k = 0, 9 do begin
      wdev = where(abs(signal(*, bref) - med) lt 3. * dev)
      med = median(signal(wdev, bref))
      dev = sqrt(mean((signal(wdev, bref) - med)^2.))
   endfor
   yrange = [-5.*dev, 8.*dev]
   otherplot, time_s, signal(*, bref) - med, charsize=1.5, yrange=yrange, title=title, $
      xtitle='time (s)', ytitle='brightness + offset'
   otherplot, /over, psym=4, time_save, array_drifts(*, bref) - 3. * dev, color=10
   otherplot, /over, time_s, drifts(*, bref) - 3. * dev, color=100
   for i = 0, nscans - 2 do begin
      otherplot, /over, color=200, lstyle=5., time_s(ind_scans(i+1)-1) + fltarr(2), yrange
      otherplot, /over, color=200, lstyle=5., time_s(ind_scans(i+1)) + fltarr(2), yrange
   endfor
   if nsubscans le 30 then otherplot, /over, time_s, dev * (13.8 * flag_nza - 10.), color=200
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')

   otherplot, time_s, signal(*, bref) - med, charsize=1.5, yrange=yrange, $
      title='zoom', xtitle='time (s)', ytitle='brightness + offset', $
      xrange=time_s([ind_subscans(0), ind_subscans(min([4, nsubscans]))-1])
   otherplot, /over, psym=4, time_save, array_drifts(*, bref) - 3. * dev, color=10
   otherplot, /over, time_s, drifts(*, bref) - 3. * dev, color=100
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif



print, '              drifts amplitude / noise (median, min and max) = ' $
   + chain(median(amp_drifts / maxnoise), 2) + ' ; ' $
   + chain(min(amp_drifts / maxnoise, /nan), 2) + ' to ' + chain(max(amp_drifts / maxnoise, /nan), 2) + crs



;;; update signal:
signal -= drifts
save, file=file_signal, signal, /xdr


;;; visualization of drifts:
if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   restore, file_maps
   proj_map_artemis, map_nodrifts, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans, $
      ind_subscans=ind_subscans
   title = 'map of iteration ' + chain(iter_tot) + ' drifts'
   if keyword_set(detect_gl) then title = title + ' (+ glitches)'
   map_drifts = map_signal - map_nodrifts
   disp_ima, map_drifts, win=3, title=title, weightmap=map_weight, noise_level=median(maxnoise)
   map_ratio = fltarr(dimx, dimy)
   w = where(map_signal gt median(maxnoise))
   map_ratio(w) = abs(map_drifts(w)) / map_signal(w)
   disp_ima, map_ratio, win=4, title='ratio map of iteration ' + chain(iter_tot) + ' |drifts| / signal', $
      weightmap=map_weight, noise_level=median(maxnoise), max_map=min([max(map_ratio), 0.15])
   if keyword_set(debug) then res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif


return, amp_drifts

end
