pro gnomonic_invert, xra, ydec, refx=refx, refy=refy


;;; inverts map coordinates using the gnomonic projection type
;;; to compute sky coordinates
;;; ref: http://lambda.gsfc.nasa.gov/product/iras/coordproj.cfm
;;;      http://mathworld.wolfram.com/GnomonicProjection.html

;;; convention xra = -RA, ydec = DEC

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



rad_deg = !dpi / 180.D0
cos_mid_dec = cos(refy * rad_deg)
sin_mid_dec = sin(refy * rad_deg)

rho = sqrt(xra^2.D0 + ydec^2.D0)
c = atan(rho * rad_deg)
ydec_sky = (asin(cos(c) * sin_mid_dec + ydec / rho * sin(c) * cos_mid_dec)) / rad_deg
xra_sky = refx + atan(xra * sin(c), rho * cos_mid_dec * cos(c) - ydec * sin_mid_dec * sin(c)) / rad_deg

xra = xra_sky
ydec = ydec_sky


end
